import contextlib
import re

import base58

from credsweeper.common.constants import ASCII
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueJfrogTokenCheck(Filter):
    """Check that candidate have a known structure JFROG token"""

    def __init__(self, config: Config = None) -> None:
        # reftkn:01:0123456789:abcdefGhijklmnoPqrstuVwxyz0
        self._pattern = re.compile(r"reftkn:\d+:\d+:[\w_/+-]+")

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        value = line_data.value
        with contextlib.suppress(Exception):
            if value.startswith("cmVmdGtuO"):
                decoded = Util.decode_base64(value, padding_safe=True, urlsafe_detect=True)
                if self._pattern.match(decoded.decode(ASCII)):
                    # identity token
                    return False
            if value.startswith("AKCp"):
                decoded = base58.b58decode(value)
                # the check only for correct size decoding
                if 54 == len(decoded):
                    # API key (deprecated) - a good integrity check solution was not found
                    return False
        return True



import io
import logging
from abc import ABC
from tarfile import TarFile
from typing import List, Optional

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class TarScanner(AbstractScanner, ABC):
    """Implements tar scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Extracts files one by one from tar archive and launches data_scan"""
        try:
            candidates = []
            with TarFile(fileobj=io.BytesIO(data_provider.data)) as tf:
                for tfi in tf.getmembers():
                    # skip directory
                    if not tfi.isreg():
                        continue
                    if FilePathExtractor.check_exclude_file(self.config, tfi.name):
                        continue
                    if 0 > recursive_limit_size - tfi.size:
                        logger.error(f"{tfi.name}: size {tfi.size}"
                                     f" is over limit {recursive_limit_size} depth:{depth}")
                        continue
                    with tf.extractfile(tfi) as f:
                        tar_content_provider = DataContentProvider(data=f.read(),
                                                                   file_path=data_provider.file_path,
                                                                   file_type=Util.get_extension(tfi.name),
                                                                   info=f"{data_provider.info}|TAR:{tfi.name}")
                        # Nevertheless, use extracted data size
                        new_limit = recursive_limit_size - len(tar_content_provider.data)
                        tar_candidates = self.recursive_scan(tar_content_provider, depth, new_limit)
                        candidates.extend(tar_candidates)
            return candidates
        except Exception as tar_exc:
            # too many exception types might be produced with broken tar
            logger.error(f"{data_provider.file_path}:{tar_exc}")
        return None



from typing import Tuple

from credsweeper.credentials.line_data import LineData


class CandidateKey:
    """Class used to identify credential candidates.

    Candidates that detected same value on same string in a same file would have identical CandidateKey
    """

    def __init__(self, line_data: LineData):
        self.path: str = line_data.path
        self.line_num: int = line_data.line_num
        self.value_start: int = line_data.value_start
        self.value_end: int = line_data.value_end
        self.key: Tuple[str, int, int, int] = (self.path, self.line_num, self.value_start, self.value_end)
        self.__line = line_data.line

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == other.key

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self) -> str:
        return f"{self.key}:{self.__line}"



import contextlib
import logging
import re
from functools import cached_property
from typing import Dict, List, Optional, Union, Set

from credsweeper import filters
from credsweeper.common.constants import RuleType, Severity, MAX_LINE_LENGTH, Confidence
from credsweeper.common.keyword_pattern import KeywordPattern
from credsweeper.config import Config
from credsweeper.filters import Filter, group
from credsweeper.filters.group import Group

logger = logging.getLogger(__name__)


class Rule:
    """Regular expression to detect some credential type.

    Class contains regular expression to detect some credential type, credential name, assumed severity
        and details on how detection should be processed

    Parameters:
        rule_name: Name displayed if rule
        rule_type: RuleType used for this rule
        severity: critical/high/medium/low
        filters: List of Filter OR _one_ filter Group that can be used to filters False detections based on rules
        patterns: Regular expressions that can be used for detection
        use_ml: Should ML work on this credential or not. If not prediction based on regular expression and filter only
        validations: List of Validation objects that can check this credential using external API
        required_substrings: Optional list of substrings. Scanner would only apply this rule if line contain at least
          one of this substrings
        min_line_len: Optional minimal line length. Scanner would only apply this rule if line is equal or longer
        usage_list: List of analyze types. There are 2 different analyze type now ("src", "doc")

    """

    # mandatory fields
    NAME = "name"
    SEVERITY = "severity"
    CONFIDENCE = "confidence"
    TYPE = "type"
    VALUES = "values"
    MIN_LINE_LEN = "min_line_len"
    TARGET = "target"

    mandatory_fields = {NAME, SEVERITY, CONFIDENCE, TYPE, VALUES, MIN_LINE_LEN, TARGET}

    # auxiliary fields
    FILTER_TYPE = "filter_type"
    USE_ML = "use_ml"
    REQUIRED_SUBSTRINGS = "required_substrings"
    REQUIRED_REGEX = "required_regex"

    all_fields = mandatory_fields | {FILTER_TYPE, USE_ML, REQUIRED_SUBSTRINGS, REQUIRED_REGEX}

    def __init__(self, config: Config, rule_dict: Dict) -> None:
        self.config = config
        self._verify_rule_config(rule_dict)
        # mandatory fields
        self.__rule_name = str(rule_dict[Rule.NAME])
        if severity := Severity.get(rule_dict[Rule.SEVERITY]):
            self.__severity = severity
        else:
            self._malformed_rule_error(rule_dict, Rule.SEVERITY)
        if confidence := Confidence.get(rule_dict[Rule.CONFIDENCE]):
            self.__confidence = confidence
        else:
            self._malformed_rule_error(rule_dict, Rule.CONFIDENCE)
        if rule_type := getattr(RuleType, str(rule_dict[Rule.TYPE]).upper(), None):
            self.__rule_type: RuleType = rule_type
        else:
            self._malformed_rule_error(rule_dict, Rule.TYPE)
        self.__patterns = self._init_patterns(rule_dict[Rule.VALUES])
        self.__target: List[str] = rule_dict.get(Rule.TARGET, [])
        if not self.__target or set(self.__target).difference({"code", "doc"}):
            self._malformed_rule_error(rule_dict, Rule.TARGET)
        # auxiliary fields
        self.__filters = self._init_filters(rule_dict.get(Rule.FILTER_TYPE, []))
        self.__use_ml = bool(rule_dict.get(Rule.USE_ML))
        self.__required_substrings = set(i.strip().lower() for i in rule_dict.get(Rule.REQUIRED_SUBSTRINGS, []))
        self.__has_required_substrings = bool(self.__required_substrings)
        required_regex = rule_dict.get(Rule.REQUIRED_REGEX)
        if required_regex and not isinstance(required_regex, str):
            self._malformed_rule_error(rule_dict, Rule.REQUIRED_REGEX)
        self.__required_regex = re.compile(required_regex) if required_regex else None
        self.__min_line_len = int(rule_dict.get(Rule.MIN_LINE_LEN, MAX_LINE_LENGTH))

    def _malformed_rule_error(self, rule_dict: Dict, field: str):
        raise ValueError(f"Malformed rule '{self.__rule_name}'."
                         f" field '{field}' has invalid value"
                         f" '{rule_dict.get(field)}'")

    @cached_property
    def rule_name(self) -> str:
        """rule_name getter"""
        return self.__rule_name

    @cached_property
    def rule_type(self) -> RuleType:
        """rule_type getter"""
        return self.__rule_type

    @cached_property
    def severity(self) -> Severity:
        """severity getter"""
        return self.__severity

    @cached_property
    def confidence(self) -> Confidence:
        """confidence getter"""
        return self.__confidence

    @cached_property
    def filters(self) -> List[Filter]:
        """filters getter"""
        return self.__filters

    @staticmethod
    def _get_arg(arg: str) -> Union[int, float, str]:
        """Transform given string value to int, then float. In worst case - returns str"""
        with contextlib.suppress(Exception):
            return int(arg)
        with contextlib.suppress(Exception):
            return float(arg)
        return str(arg)

    def _init_filters(self, filter_type: Union[None, str, List[str]]) -> List[Filter]:
        """
            filter_type: str - applies Group of filter
                         list - creates specific set of Filters
        """
        _filters: List[Filter] = []
        if isinstance(filter_type, str):
            # when string passed - (Group) of filters is applied
            filter_group = getattr(group, filter_type, None)
            if isinstance(filter_group, type) and issubclass(filter_group, Group):
                return filter_group(self.config).filters  # type: ignore
        elif isinstance(filter_type, list):
            # list type means - list of (Filter)s is applied
            for i in filter_type:
                if '(' in i and ')' in i:
                    left_pos = i.find('(')
                    filter_parameters = [self._get_arg(x.strip()) for x in i[left_pos + 1:i.find(')')].split(',')]
                    filter_name = i[:left_pos].strip()
                else:
                    filter_parameters = None
                    filter_name = i
                _filter = getattr(filters, filter_name, None)
                if isinstance(_filter, type) and issubclass(_filter, Filter):
                    if filter_parameters:
                        _filters.append(_filter(self.config, *filter_parameters))
                    else:
                        _filters.append(_filter(self.config))
                else:
                    break
            else:
                return _filters
        raise ValueError(f"Malformed rule '{self.__rule_name}'."
                         f" field '{Rule.FILTER_TYPE}' has invalid value"
                         f" '{filter_type}'")

    def _init_patterns(self, _values: List[str]) -> List[re.Pattern]:
        """Get pattern values for rule object.

        Set the pattern value attribute of the rule object based on the passed values.
        So, if the received rule type corresponds to the RuleType.KEYWORD type,
        the "patterns" attribute is assigned the value of template keyword regex
        with the corresponding value. Otherwise, if the received rule type corresponds
        to the RuleType.PATTERN, RuleType.MULTI or RuleType.PEM_KEY types, the "patterns" attribute is
        assigned the compile regex ov received value

        Args:
            _values: regular expressions

        """
        _patterns: List[re.Pattern] = []
        if RuleType.KEYWORD == self.rule_type and 0 < len(_values):
            for value in _values:
                _pattern = KeywordPattern.get_keyword_pattern(value)
                _patterns.append(_pattern)
            return _patterns
        elif RuleType.MULTI == self.rule_type and 2 == len(_values) \
                or self.rule_type in (RuleType.PATTERN, RuleType.PEM_KEY) and 0 < len(_values):
            for value in _values:
                _patterns.append(re.compile(value))
            if RuleType.PEM_KEY == self.rule_type and 1 < len(_values):
                logger.warning(f"Rule {self.rule_name} has extra patterns. Only single pattern supported.")
            elif RuleType.MULTI == self.rule_type and 2 < len(_values):
                logger.warning(f"Rule {self.rule_name} has extra patterns. Only two patterns supported.")
            return _patterns
        raise ValueError(f"Malformed rule config file. Rule '{self.rule_name}' type '{self.rule_type}' is invalid.")

    @cached_property
    def patterns(self) -> List[re.Pattern]:
        """patterns getter"""
        return self.__patterns

    @cached_property
    def use_ml(self) -> bool:
        """use_ml getter"""
        return self.__use_ml

    @staticmethod
    def _verify_rule_config(rule_config: Dict) -> None:
        """Checks all mandatory fields and wrong names

        Args:
            rule_config: dictionary loaded from the config file

        Raises:
            ValueError if missing fields is present

        """
        if missing_fields := Rule.mandatory_fields.difference(rule_config.keys()):
            raise ValueError(f"Malformed rule config file. Contain rule with missing fields: {missing_fields}.")
        if extra_fields := set(rule_config.keys()).difference(Rule.all_fields):
            raise ValueError(f"Malformed rule config file. Extra fields: {extra_fields}.")

    @cached_property
    def required_substrings(self) -> Set[str]:
        """required_substrings getter"""
        return self.__required_substrings

    @cached_property
    def has_required_substrings(self) -> bool:
        """has_required_substrings getter for speedup"""
        return self.__has_required_substrings

    @cached_property
    def required_regex(self) -> Optional[re.Pattern]:
        """required_regex getter"""
        return self.__required_regex

    @cached_property
    def min_line_len(self) -> int:
        """min_line_len getter"""
        return self.__min_line_len

    @cached_property
    def target(self) -> List[str]:
        """target getter"""
        return self.__target



from credsweeper.common.constants import CHUNK_SIZE
from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.feature import Feature
from credsweeper.utils import Util


class HasHtmlTag(Feature):
    """Feature is true if line has HTML tags (HTML file)."""

    def __init__(self) -> None:
        super().__init__()
        self.words = [
            '< img', '<img', '< script', '<script', '< p', '<p', '< link', '<link', '< meta', '<meta', '< a', '<a'
        ]

    def extract(self, candidate: Candidate) -> bool:
        subtext = Util.subtext(candidate.line_data_list[0].line, candidate.line_data_list[0].value_start, CHUNK_SIZE)
        candidate_line_data_list_0_line_lower = subtext.lower()
        if '<' not in candidate_line_data_list_0_line_lower:
            # early check
            return False
        if self.any_word_in_(candidate_line_data_list_0_line_lower):
            return True
        if "/>" in candidate_line_data_list_0_line_lower or "</" in candidate_line_data_list_0_line_lower:
            # possible closed tag
            return True
        return False



import io
import logging
from abc import ABC
from typing import List, Optional

import docx
from docx.document import Document
from docx.oxml import CT_P, CT_Tbl, CT_SectPr, CT_TcPr
from docx.section import Section, _Header, _Footer
from docx.table import _Cell, Table
from docx.text.paragraph import Paragraph
from lxml.etree import _Element

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider
from credsweeper.file_handler.string_content_provider import StringContentProvider

logger = logging.getLogger(__name__)


class DocxScanner(AbstractScanner, ABC):
    """Implements docx scanning"""

    @staticmethod
    def _iter_block_items(block):
        if isinstance(block, Paragraph):
            yield block
            return
        if isinstance(block, (_Header, _Footer)):
            for table in block.tables:
                for row in table.rows:
                    for cell in row.cells:
                        yield from DocxScanner._iter_block_items(cell)
            for paragraph in block.paragraphs:
                yield paragraph
            return
        elif isinstance(block, Document):
            parent_elm = block.element.body
        elif isinstance(block, Section):
            yield from DocxScanner._iter_block_items(block.header)
            yield from DocxScanner._iter_block_items(block.footer)
            return
        elif isinstance(block, _Cell):
            parent_elm = block._tc
        else:
            raise ValueError(f"unrecognised:{type(block)}")

        for child in parent_elm.iterchildren():
            if isinstance(child, CT_P):
                yield Paragraph(child, block)
            elif isinstance(child, CT_Tbl):
                table = Table(child, block)
                for row in table.rows:
                    for cell in row.cells:
                        yield from DocxScanner._iter_block_items(cell)
            elif isinstance(child, (CT_TcPr, CT_SectPr)):
                # config
                pass
            elif isinstance(child, _Element):
                yield child
            else:
                logger.warning(f"Unknown:{type(child)}")

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan DOCX text with splitting by lines"""
        try:
            docx_lines: List[str] = []

            doc = docx.Document(io.BytesIO(data_provider.data))
            for block in self._iter_block_items(doc):
                if block.text:
                    docx_lines.append(block.text)

            header_lines_set = set()
            footer_lines_set = set()
            for section in doc.sections:
                for header in [section.first_page_header, section.even_page_header, section.header]:
                    for block in self._iter_block_items(header):
                        if block.text:
                            header_lines_set.add(block.text)
                for footer in [section.first_page_footer, section.even_page_footer, section.footer]:
                    for block in self._iter_block_items(footer):
                        if block.text:
                            footer_lines_set.add(block.text)
            docx_lines.extend(sorted(list(header_lines_set)))
            docx_lines.extend(sorted(list(footer_lines_set)))

            string_data_provider = StringContentProvider(lines=docx_lines,
                                                         file_path=data_provider.file_path,
                                                         file_type=data_provider.file_type,
                                                         info=f"{data_provider.info}|DOCX")
            docx_candidates = self.scanner.scan(string_data_provider)
            return docx_candidates

        except Exception as docx_exc:
            logger.error(f"{data_provider.file_path}:{docx_exc}")
        return None



import logging
import multiprocessing
import signal
from pathlib import Path
from typing import Any, List, Optional, Union, Dict, Sequence, Tuple

import pandas as pd
from colorama import Style

# Directory of credsweeper sources MUST be placed before imports to avoid circular import error
APP_PATH = Path(__file__).resolve().parent

from credsweeper.common.constants import Severity, ThresholdPreset, DiffRowType
from credsweeper.config import Config
from credsweeper.credentials import Candidate, CredentialManager, CandidateKey
from credsweeper.deep_scanner.deep_scanner import DeepScanner
from credsweeper.file_handler.diff_content_provider import DiffContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.file_handler.abstract_provider import AbstractProvider
from credsweeper.file_handler.text_content_provider import TextContentProvider
from credsweeper.scanner import Scanner
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class CredSweeper:
    """Advanced credential analyzer base class.

    Parameters:
        credential_manager: CredSweeper credential manager object
        scanner: CredSweeper scanner object
        pool_count: number of pools used to run multiprocessing scanning
        config: dictionary variable, stores analyzer features
        json_filename: string variable, credential candidates export filename

    """

    def __init__(self,
                 rule_path: Union[None, str, Path] = None,
                 config_path: Optional[str] = None,
                 json_filename: Union[None, str, Path] = None,
                 xlsx_filename: Union[None, str, Path] = None,
                 color: bool = False,
                 hashed: bool = False,
                 subtext: bool = False,
                 sort_output: bool = False,
                 use_filters: bool = True,
                 pool_count: int = 1,
                 ml_batch_size: Optional[int] = None,
                 ml_threshold: Union[float, ThresholdPreset] = ThresholdPreset.medium,
                 ml_config: Union[None, str, Path] = None,
                 ml_model: Union[None, str, Path] = None,
                 ml_providers: Optional[str] = None,
                 find_by_ext: bool = False,
                 depth: int = 0,
                 doc: bool = False,
                 severity: Union[Severity, str] = Severity.INFO,
                 size_limit: Optional[str] = None,
                 exclude_lines: Optional[List[str]] = None,
                 exclude_values: Optional[List[str]] = None,
                 thrifty: bool = False,
                 log_level: Optional[str] = None) -> None:
        """Initialize Advanced credential scanner.

        Args:
            rule_path: optional str variable, path of rule config file
                validation was the grained candidate model on machine learning
            config_path: optional str variable, path of CredSweeper config file
                default built-in config is used if None
            json_filename: optional string variable, path to save result to json
            xlsx_filename: optional string variable, path to save result to xlsx
            color: print results to stdout with colorization
            hashed: use hash of line, value and variable instead plain text
            subtext: use subtext of line near variable-value like it performed in ML
            use_filters: boolean variable, specifying the need of rule filters
            pool_count: int value, number of parallel processes to use
            ml_batch_size: int value, size of the batch for model inference
            ml_threshold: float or string value to specify threshold for the ml model
            ml_config: str or Path to set custom config of ml model
            ml_model: str or Path to set custom ml model
            ml_providers: str - comma separated list with providers
            find_by_ext: boolean - files will be reported by extension
            depth: int - how deep container files will be scanned
            doc: boolean - document-specific scanning
            severity: Severity - minimum severity level of rule
            size_limit: optional string integer or human-readable format to skip oversize files
            exclude_lines: lines to omit in scan. Will be added to the lines already in config
            exclude_values: values to omit in scan. Will be added to the values already in config
            thrifty: free provider resources after scan to reduce memory consumption
            log_level: str - level for pool initializer according logging levels (UPPERCASE)

        """
        self.pool_count: int = int(pool_count) if int(pool_count) > 1 else 1
        if not (_severity := Severity.get(severity)):
            raise RuntimeError(f"Severity level provided: {severity}"
                               f" -- must be one of: {' | '.join([i.value for i in Severity])}")
        config_dict = self._get_config_dict(config_path=config_path,
                                            use_filters=use_filters,
                                            find_by_ext=find_by_ext,
                                            depth=depth,
                                            doc=doc,
                                            severity=_severity,
                                            size_limit=size_limit,
                                            exclude_lines=exclude_lines,
                                            exclude_values=exclude_values)
        self.config = Config(config_dict)
        self.scanner = Scanner(self.config, rule_path)
        self.deep_scanner = DeepScanner(self.config, self.scanner)
        self.credential_manager = CredentialManager()
        self.json_filename: Union[None, str, Path] = json_filename
        self.xlsx_filename: Union[None, str, Path] = xlsx_filename
        self.color = color
        self.hashed = hashed
        self.subtext = subtext
        self.sort_output = sort_output
        self.ml_batch_size = ml_batch_size if ml_batch_size and 0 < ml_batch_size else 16
        self.ml_threshold = ml_threshold
        self.ml_config = ml_config
        self.ml_model = ml_model
        self.ml_providers = ml_providers
        self.ml_validator = None
        self.__thrifty = thrifty
        self.__log_level = log_level

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @staticmethod
    def _get_config_path(config_path: Optional[str]) -> Path:
        if config_path:
            return Path(config_path)
        else:
            return APP_PATH / "secret" / "config.json"

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def _get_config_dict(
            self,  #
            config_path: Optional[str],  #
            use_filters: bool,  #
            find_by_ext: bool,  #
            depth: int,  #
            doc: bool,  #
            severity: Severity,  #
            size_limit: Optional[str],  #
            exclude_lines: Optional[List[str]],  #
            exclude_values: Optional[List[str]]) -> Dict[str, Any]:
        config_dict = Util.json_load(self._get_config_path(config_path))
        config_dict["use_filters"] = use_filters
        config_dict["find_by_ext"] = find_by_ext
        config_dict["size_limit"] = size_limit
        config_dict["depth"] = depth
        config_dict["doc"] = doc
        config_dict["severity"] = severity.value

        if exclude_lines is not None:
            config_dict["exclude"]["lines"] = config_dict["exclude"].get("lines", []) + exclude_lines
        if exclude_values is not None:
            config_dict["exclude"]["values"] = config_dict["exclude"].get("values", []) + exclude_values

        return config_dict  # type: ignore

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def _use_ml_validation(self) -> bool:
        if isinstance(self.ml_threshold, (float, int)) and 0 >= self.ml_threshold:
            logger.info("ML validation is disabled")
            return False
        if not self.credential_manager.candidates:
            logger.info("Skip ML validation because no candidates were found")
            return False
        for i in self.credential_manager.candidates:
            if i.use_ml:
                # any() or all() is not used to speedup
                return True
        logger.info("Skip ML validation because no candidates support it")
        return False

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    # the import cannot be done on top due
    # TypeError: cannot pickle 'onnxruntime.capi.onnxruntime_pybind11_state.InferenceSession' object
    from credsweeper.ml_model import MlValidator

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @property
    def ml_validator(self) -> MlValidator:
        """ml_validator getter"""
        from credsweeper.ml_model import MlValidator
        if not self.__ml_validator:
            self.__ml_validator: MlValidator = MlValidator(
                threshold=self.ml_threshold,  #
                ml_config=self.ml_config,  #
                ml_model=self.ml_model,  #
                ml_providers=self.ml_providers,  #
            )
        assert self.__ml_validator, "self.__ml_validator was not initialized"
        return self.__ml_validator

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @ml_validator.setter
    def ml_validator(self, _ml_validator: Optional[MlValidator]) -> None:
        """ml_validator setter"""
        self.__ml_validator = _ml_validator

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @staticmethod
    def pool_initializer(log_kwargs) -> None:
        """Ignore SIGINT in child processes."""
        logging.basicConfig(**log_kwargs)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @property
    def config(self) -> Config:
        """config getter"""
        return self.__config

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @config.setter
    def config(self, config: Config) -> None:
        """config setter"""
        self.__config = config

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def run(self, content_provider: AbstractProvider) -> int:
        """Run an analysis of 'content_provider' object.

        Args:
            content_provider: path objects to scan

        """
        _empty_list: Sequence[Union[DiffContentProvider, TextContentProvider]] = []
        file_extractors: Sequence[Union[DiffContentProvider, TextContentProvider]] = \
            content_provider.get_scannable_files(self.config) if content_provider else _empty_list
        logger.info(f"Start Scanner for {len(file_extractors)} providers")
        self.scan(file_extractors)
        self.post_processing()
        # PatchesProvider has the attribute. Circular import error appears with using the isinstance
        change_type = content_provider.change_type if hasattr(content_provider, "change_type") else None
        self.export_results(change_type)

        return len(self.credential_manager.get_credentials())

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def scan(self, content_providers: Sequence[Union[DiffContentProvider, TextContentProvider]]) -> None:
        """Run scanning of files from an argument "content_providers".

        Args:
            content_providers: file objects to scan

        """
        if 1 < self.pool_count:
            self.__multi_jobs_scan(content_providers)
        else:
            self.__single_job_scan(content_providers)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def __single_job_scan(self, content_providers: Sequence[Union[DiffContentProvider, TextContentProvider]]) -> None:
        """Performs scan in main thread"""
        all_cred = self.files_scan(content_providers)
        self.credential_manager.set_credentials(all_cred)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def __multi_jobs_scan(self, content_providers: Sequence[Union[DiffContentProvider, TextContentProvider]]) -> None:
        """Performs scan with multiple jobs"""
        # use this separation to satisfy YAPF formatter
        yapfix = "%(asctime)s | %(levelname)s | %(processName)s:%(threadName)s | %(filename)s:%(lineno)s | %(message)s"
        log_kwargs = {"format": yapfix}
        if isinstance(self.__log_level, str):
            # is not None
            if "SILENCE" == self.__log_level:
                logging.addLevelName(60, "SILENCE")
            log_kwargs["level"] = self.__log_level
        with multiprocessing.get_context("spawn").Pool(processes=self.pool_count,
                                                       initializer=self.pool_initializer,
                                                       initargs=(log_kwargs, )) as pool:
            try:
                for scan_results in pool.imap_unordered(self.files_scan, (content_providers[x::self.pool_count]
                                                                          for x in range(self.pool_count))):
                    for cred in scan_results:
                        self.credential_manager.add_credential(cred)
            except KeyboardInterrupt:
                pool.terminate()
                pool.join()
                raise
            pool.close()
            pool.join()

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def files_scan(
            self,  #
            content_providers: Sequence[Union[DiffContentProvider, TextContentProvider]]) -> List[Candidate]:
        """Auxiliary method for scan one sequence"""
        all_cred: List[Candidate] = []
        for provider in content_providers:
            candidates = self.file_scan(provider)
            if self.__thrifty:
                provider.free()
            all_cred.extend(candidates)
        logger.info(f"Completed: processed {len(content_providers)} providers with {len(all_cred)} candidates")
        return all_cred

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def file_scan(self, content_provider: Union[DiffContentProvider, TextContentProvider]) -> List[Candidate]:
        """Run scanning of file from 'file_provider'.

        Args:
            content_provider: content provider object to scan

        Return:
            list of credential candidates from scanned file

        """
        candidates: List[Candidate] = []
        logger.debug("Start scan file: %s %s", content_provider.file_path, content_provider.info)

        if FilePathExtractor.is_find_by_ext_file(self.config, content_provider.file_type):
            # Skip the file scanning and create fake candidate because the extension is suspicious
            dummy_candidate = Candidate.get_dummy_candidate(self.config, content_provider.file_path,
                                                            content_provider.file_type, content_provider.info,
                                                            FilePathExtractor.FIND_BY_EXT_RULE)
            candidates.append(dummy_candidate)

        else:
            if self.config.depth or self.config.doc:
                # deep scan with possible data representation
                candidates = self.deep_scanner.scan(content_provider, self.config.depth, self.config.size_limit)
            else:
                if content_provider.file_type not in self.config.exclude_containers:
                    # Regular file scanning
                    candidates = self.scanner.scan(content_provider)

        # finally return result from 'file_scan'
        return candidates

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def post_processing(self) -> None:
        """Machine learning validation for received credential candidates."""
        if purged := self.credential_manager.purge_duplicates():
            logger.info(f"Purged {purged} duplicates")
        if self._use_ml_validation():
            logger.info(f"Grouping {len(self.credential_manager.candidates)} candidates")
            new_cred_list: List[Candidate] = []
            cred_groups = self.credential_manager.group_credentials()
            ml_cred_groups: List[Tuple[CandidateKey, List[Candidate]]] = []
            for group_key, group_candidates in cred_groups.items():
                # Analyze with ML if any candidate in group require ML
                for candidate in group_candidates:
                    if candidate.use_ml:
                        ml_cred_groups.append((group_key, group_candidates))
                        break
                else:
                    # all candidates do not require ML
                    new_cred_list.extend(group_candidates)

            # prevent extra ml_validator creation if ml_cred_groups is empty
            if ml_cred_groups:
                logger.info(f"Run ML Validation for {len(ml_cred_groups)} groups")
                is_cred, probability = self.ml_validator.validate_groups(ml_cred_groups, self.ml_batch_size)
                for i, (_, group_candidates) in enumerate(ml_cred_groups):
                    for candidate in group_candidates:
                        if candidate.use_ml:
                            if is_cred[i]:
                                candidate.ml_probability = probability[i]
                                new_cred_list.append(candidate)
                        else:
                            new_cred_list.append(candidate)
            else:
                logger.info("Skipping ML validation due not applicable")

            self.credential_manager.set_credentials(new_cred_list)

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def export_results(self, change_type: Optional[DiffRowType] = None) -> None:
        """
        Save credential candidates to json file or print them to a console.

        Args:
            change_type: flag to know which file should be created for a patch
        """
        is_exported = False

        credentials = self.credential_manager.get_credentials()

        logger.info(f"Exporting {len(credentials)} credentials")

        if self.sort_output:
            credentials.sort(key=lambda x: (  #
                x.line_data_list[0].path,  #
                x.line_data_list[0].line_num,  #
                x.severity,  #
                x.rule_name,  #
                x.line_data_list[0].value_start,  #
                x.line_data_list[0].value_end  #
            ))

        if self.json_filename:
            json_path = Path(self.json_filename)
            is_exported = True
            if isinstance(change_type, DiffRowType):
                # add suffix for appropriated reports to create two files for the patch scan
                json_path = json_path.with_suffix(f".{change_type.value}{json_path.suffix}")
            Util.json_dump([credential.to_json(hashed=self.hashed, subtext=self.subtext) for credential in credentials],
                           file_path=json_path)

        if self.xlsx_filename:
            is_exported = True
            data_list = []
            for credential in credentials:
                data_list.extend(credential.to_dict_list(hashed=self.hashed, subtext=self.subtext))
            df = pd.DataFrame(data=data_list)
            if isinstance(change_type, DiffRowType):
                if Path(self.xlsx_filename).exists():
                    with pd.ExcelWriter(self.xlsx_filename, mode='a', engine="openpyxl",
                                        if_sheet_exists="replace") as writer:
                        df.to_excel(writer, sheet_name=change_type.value, index=False)
                else:
                    df.to_excel(self.xlsx_filename, sheet_name=change_type.value, index=False)
            else:
                df.to_excel(self.xlsx_filename, sheet_name="report", index=False)

        if self.color:
            is_exported = True
            for credential in credentials:
                for line_data in credential.line_data_list:
                    # bright rule name and path or info
                    print(Style.BRIGHT + credential.rule_name +
                          f" {line_data.info or line_data.path}:{line_data.line_num} {credential.ml_probability}" +
                          Style.RESET_ALL)
                    print(line_data.get_colored_line(hashed=self.hashed, subtext=self.subtext))

        if is_exported is False:
            for credential in credentials:
                print(credential.to_str(hashed=self.hashed, subtext=self.subtext))



from typing import List

import numpy as np

from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInVariable(WordIn):
    """Feature returns array of words matching in variable"""

    def __init__(self, words: List[str]) -> None:
        """Feature is true if candidate value contains at least one predefined word.

        Args:
            words: list of predefined words - MUST BE IN LOWER CASE

        """
        super().__init__(words)

    def extract(self, candidate: Candidate) -> np.ndarray:
        """Returns array of matching words for first line"""
        if variable := candidate.line_data_list[0].variable:
            return self.word_in_str(variable.lower())
        else:
            return np.zeros(shape=[self.dimension], dtype=np.int8)



import logging
from abc import ABC
from typing import List, Optional

import cryptography.hazmat.primitives.serialization.pkcs12

from credsweeper.credentials import Candidate
from credsweeper.deep_scanner.abstract_scanner import AbstractScanner
from credsweeper.file_handler.data_content_provider import DataContentProvider

logger = logging.getLogger(__name__)


class Pkcs12Scanner(AbstractScanner, ABC):
    """Implements pkcs12 scanning"""

    def data_scan(
            self,  #
            data_provider: DataContentProvider,  #
            depth: int,  #
            recursive_limit_size: int) -> Optional[List[Candidate]]:
        """Tries to scan PKCS12 to open with standard password"""
        candidates = []
        for pw_probe in self.config.bruteforce_list:
            try:
                (private_key, certificate, additional_certificates) \
                    = cryptography.hazmat.primitives.serialization.pkcs12.load_key_and_certificates(data_provider.data,
                                                                                                    pw_probe.encode())
                # the password probe has passed, it will be the value
                value = pw_probe or "<EMPTY PASSWORD>"
                info = (f"{data_provider.info}|PKCS12:"
                        f"'{value}' {'sensitive data' if private_key else 'default password'}")
                candidate = Candidate.get_dummy_candidate(
                    self.config,  #
                    data_provider.file_path,  #
                    data_provider.file_type,  #
                    info,  #
                    "PKCS12")
                candidate.line_data_list[0].line = f"'{value}' is the password"
                candidate.line_data_list[0].value = value
                candidate.line_data_list[0].value_start = 1
                candidate.line_data_list[0].value_end = 1 + len(candidate.line_data_list[0].value)
                candidates.append(candidate)
                break
            except Exception as pkcs_exc:
                logger.debug(f"{data_provider.file_path}:{pw_probe}:{pkcs_exc}")
        return candidates



import statistics
from typing import Tuple


class HopStat:
    """Statistical check distances between symbols sequence in a value on keyboard layout"""

    KEYBOARD = (  #
        "`1234567890-=",  #
        "\0qwertyuiop[]\\",  #
        "\0\0asdfghjkl;'",  #
        "\0\0zxcvbnm,./",  #
    )
    TRANSLATION = str.maketrans({
        '~': '`',
        '!': '1',
        '@': '2',
        '#': '3',
        '$': '4',
        '%': '5',
        '^': '6',
        '&': '7',
        '*': '8',
        '(': '9',
        ')': '0',
        '_': '-',
        '+': '=',
        '{': '[',
        '}': ']',
        '|': '\\',
        ':': ';',
        '"': "'",
        '<': ',',
        '>': '.',
        '?': '/',
    })

    def __init__(self):
        self.__hop_dict = dict()
        base = ''.join(x for x in HopStat.KEYBOARD)
        for a in (x for x in base if '\0' != x):
            for b in (x for x in base if '\0' != x):
                if (b, a) in self.__hop_dict:
                    self.__hop_dict[(a, b)] = self.__hop_dict[(b, a)]
                    continue
                if a == b:
                    self.__hop_dict[(a, b)] = 0
                else:
                    x_a, y_a, z_a = self.__get_xyz(a)
                    x_b, y_b, z_b = self.__get_xyz(b)
                    d = (abs(x_a - x_b) + abs(y_a - y_b) + abs(z_a - z_b)) // 2
                    self.__hop_dict[(a, b)] = d

    @staticmethod
    def __get_xyz(c: str) -> Tuple[int, int, int]:
        """Returns axial coordinates of a char on keyboad qwerty"""
        x = y = z = 0
        for i in range(len(HopStat.KEYBOARD)):
            x = HopStat.KEYBOARD[i].find(c)
            if 0 <= x:
                z = i
                x = x - (i // 2)
                y = -(z + x)
                break
        return x, y, z

    def stat(self, value: str) -> Tuple[float, float]:
        """Calculates statistical distances between given symbols

        Args:
            value: string based on initial alphabet

        Return:
            Average distance, deviation or exception if a value is not in initial alphabet

        """
        hops = []
        value = value.lower().translate(HopStat.TRANSLATION)
        for a, b in zip(value[:-1], value[1:]):
            hop = self.__hop_dict.get((a, b))
            if hop is None:
                raise ValueError(f"Unknown char '{a}' or '{b}'")
            hops.append(hop)
        avg = statistics.mean(hops)
        dev = statistics.stdev(hops, avg)
        return avg, dev



import math

from credsweeper.common.constants import Chars, ENTROPY_LIMIT_BASE64
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueEntropyBase64Check(Filter):
    """Check that candidate have Shanon Entropy > 3 (for HEX_CHARS or BASE36_CHARS) or > 4.5 (for BASE64_CHARS)."""

    # If the value size is less than this value the entropy evaluation gives an imprecise result
    min_length = 12

    def __init__(self, config: Config = None) -> None:
        pass

    @staticmethod
    def get_min_data_entropy(x: int) -> float:
        """Returns minimal average entropy for size of random data. Precalculated round data is applied for speedup"""
        if 18 == x:
            y = 3.8
        elif 20 == x:
            y = 3.9
        elif 24 == x:
            y = 4.1
        elif 32 == x:
            y = 4.4
        elif ValueEntropyBase64Check.min_length <= x < 35:
            # logarithm base 2 - slow, but precise. Approximation does not exceed stdev
            y = 0.77 * math.log2(x) + 0.62
        elif 35 <= x < 60:
            y = ENTROPY_LIMIT_BASE64
        elif 60 <= x:
            # the entropy grows slowly after 60
            y = 5.0
        else:
            y = 0
        return y

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if '-' in line_data.value or '_' in line_data.value:
            entropy = Util.get_shannon_entropy(line_data.value, Chars.BASE64URL_CHARS.value)
        else:
            entropy = Util.get_shannon_entropy(line_data.value, Chars.BASE64STD_CHARS.value)
        min_entropy = ValueEntropyBase64Check.get_min_data_entropy(len(line_data.value))
        return min_entropy > entropy or 0 == min_entropy



import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueMethodCheck(Filter):
    """Check if potential candidate value is a function.

    Check if potential candidate value is a function by looking for '(', ')' or 'function' sub-strings in it
    """

    PATTERN = re.compile(r"^[~.\->:0-9A-Za-z_]+\(.*\)")

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.is_well_quoted_value:
            return False
        if "function" in line_data.value or self.PATTERN.search(line_data.value):
            return True
        return False



from pathlib import Path
from typing import List, Any

import numpy as np

from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class WordInPath(WordIn):
    """Categorical feature that corresponds to words in path (POSIX, lowercase)"""

    def __init__(self, words: List[str]) -> None:
        """WordInPath constructor

        Args:
            words: list of predefined words - MUST BE IN LOWER CASE & POSIX

        """
        super().__init__(words)

    def __call__(self, candidates: List[Candidate]) -> np.ndarray:
        # actually there must be one path because the candidates are grouped before
        if path := candidates[0].line_data_list[0].path:
            posix_lower_path = Path(path).as_posix().lower()
            return self.word_in_str(posix_lower_path)
        else:
            return np.array([np.zeros(shape=[self.dimension], dtype=np.int8)])

    def extract(self, candidate: Candidate) -> Any:
        raise NotImplementedError



import contextlib

from credsweeper.common.constants import Chars
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.filters.value_entropy_base64_check import ValueEntropyBase64Check
from credsweeper.utils import Util


class ValueDiscordBotCheck(Filter):
    """Discord bot Token"""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        with contextlib.suppress(Exception):
            # . must be in value according regex
            dot_separator_index = line_data.value.index('.')
            id_part = line_data.value[:dot_separator_index]
            discord_id = int(Util.decode_base64(id_part, padding_safe=True, urlsafe_detect=True))
            entropy_part = line_data.value[dot_separator_index:]
            entropy = Util.get_shannon_entropy(entropy_part, Chars.BASE64URL_CHARS.value)
            min_entropy = ValueEntropyBase64Check.get_min_data_entropy(len(entropy_part))
            if 1000 <= discord_id and min_entropy <= entropy:
                return False
        return True



from credsweeper.common.constants import GroupType
from credsweeper.config import Config
from credsweeper.filters import ValueCoupleKeywordCheck, ValueNotPartEncodedCheck, \
    ValueBase64DataCheck, ValueEntropyBase64Check, ValuePatternCheck, ValueNumberCheck, ValueTokenBase64Check, \
    ValueBase64PartCheck
from credsweeper.filters.group import Group


class WeirdBase64Token(Group):
    """Structured Token"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.DEFAULT)
        self.filters = [
            ValueCoupleKeywordCheck(),
            ValueNumberCheck(),
            ValueBase64DataCheck(),
            ValueTokenBase64Check(),
            ValueEntropyBase64Check(),
            ValuePatternCheck(config),
            ValueNotPartEncodedCheck(),
            ValueBase64PartCheck(),
        ]



from credsweeper.common.constants import GroupType
from credsweeper.config import Config
from credsweeper.filters import ValueCoupleKeywordCheck, ValuePatternCheck, ValueNumberCheck, ValueEntropyBase36Check, \
    ValueTokenBase36Check
from credsweeper.filters.group import Group


class WeirdBase36Token(Group):
    """Structured Token"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.DEFAULT)
        self.filters = [
            ValueCoupleKeywordCheck(),
            ValuePatternCheck(config),
            ValueNumberCheck(),
            ValueTokenBase36Check(),
            ValueEntropyBase36Check()
        ]



import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class LineUUEPartCheck(Filter):
    """Checks that line is not a part of UU encoding only for maximal line"""
    uue_string = re.compile(r"^M[!-`]{60}$")

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if not line_data.line:
            return True
        if 61 != target.line_len:
            return False
        line = target.line
        if LineUUEPartCheck.uue_string.match(line):
            # to be sure - check two lines: before and/or after
            if 0 < line_data.line_pos:
                previous_line = target.lines[line_data.line_pos - 1]
                if LineUUEPartCheck.uue_string.match(previous_line):
                    return True

            if len(target.lines) > 1 + line_data.line_pos:
                next_line = target.lines[line_data.line_pos + 1]
                if LineUUEPartCheck.uue_string.match(next_line):
                    return True

        return False



from credsweeper.common.constants import GroupType
from credsweeper.config import Config
from credsweeper.filters import (ValueAllowlistCheck, ValueArrayDictionaryCheck, ValueBlocklistCheck,
                                 ValueCamelCaseCheck, ValueDictionaryValueLengthCheck, ValueFilePathCheck,
                                 ValueFirstWordCheck, ValueLastWordCheck, ValueMethodCheck, ValueNotAllowedPatternCheck,
                                 ValuePatternCheck, ValueStringTypeCheck, ValueTokenCheck)
from credsweeper.filters.group import Group


class UrlCredentialsGroup(Group):
    """UrlCredentialsGroup"""

    def __init__(self, config: Config) -> None:
        """URL credentials group class.

        Similar to PasswordKeyword, but exclude all checks dependent on the variable name, as URL credentials have no
        explicitly defined variable
        """
        super().__init__(config, GroupType.DEFAULT)
        self.filters = [
            ValueAllowlistCheck(),
            ValueArrayDictionaryCheck(),
            ValueBlocklistCheck(),
            ValueCamelCaseCheck(),
            ValueFilePathCheck(),
            ValueFirstWordCheck(),
            ValueLastWordCheck(),
            ValueMethodCheck(),
            ValueStringTypeCheck(config),
            ValueNotAllowedPatternCheck(),
            ValueTokenCheck(),
            ValueDictionaryValueLengthCheck(min_len=4, max_len=80),
            ValuePatternCheck(pattern_len=config.pattern_len)
        ]



from credsweeper.common.constants import GroupType
from credsweeper.config import Config
from credsweeper.filters.group import Group


class GeneralPattern(Group):
    """GeneralPattern"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.PATTERN)



import hashlib
import logging
from pathlib import Path
from typing import List, Tuple, Union, Optional, Dict

import numpy as np
import onnxruntime as ort

import credsweeper.ml_model.features as features
from credsweeper.common.constants import ThresholdPreset, ML_HUNK
from credsweeper.credentials import Candidate, CandidateKey
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class MlValidator:
    """ML validation class"""
    MAX_LEN = 2 * ML_HUNK  # for whole line limit
    # used for initial fill
    ZERO_CHAR = '\x00'
    # applied for unknown characters
    FAKE_CHAR = '\x01'

    def __init__(
            self,  #
            threshold: Union[float, ThresholdPreset],  #
            ml_config: Union[None, str, Path] = None,  #
            ml_model: Union[None, str, Path] = None,  #
            ml_providers: Optional[str] = None) -> None:
        """Init

        Args:
            threshold: decision threshold
            ml_config: path to ml config
            ml_model: path to ml model
            ml_providers: coma separated list of providers https://onnxruntime.ai/docs/execution-providers/
        """
        dir_path = Path(__file__).parent

        if ml_config:
            ml_config_path = Path(ml_config)
        else:
            ml_config_path = dir_path / "ml_config.json"
        with open(ml_config_path, "rb") as f:
            md5_config = hashlib.md5(f.read()).hexdigest()

        if ml_model:
            ml_model_path = Path(ml_model)
        else:
            ml_model_path = dir_path / "ml_model.onnx"
        with open(ml_model_path, "rb") as f:
            md5_model = hashlib.md5(f.read()).hexdigest()

        if ml_providers:
            providers = ml_providers.split(',')
        else:
            providers = ["CPUExecutionProvider"]
        self.model_session = ort.InferenceSession(ml_model_path, providers=providers)

        model_config = Util.json_load(ml_config_path)
        if isinstance(threshold, float):
            self.threshold = threshold
        elif isinstance(threshold, ThresholdPreset) and "thresholds" in model_config:
            self.threshold = model_config["thresholds"][threshold.value]
        else:
            self.threshold = 0.5

        char_set = set(model_config["char_set"])
        if len(char_set) != len(model_config["char_set"]):
            logger.warning('Duplicated symbols in "char_set"?')
        if self.ZERO_CHAR in char_set or self.FAKE_CHAR in char_set:
            raise ValueError(f'Unacceptable symbols 0x00 or 0x01 in "char_set"={char_set}')
        self.char_dict = {self.ZERO_CHAR: 0, self.FAKE_CHAR: 1}
        self.char_dict.update({
            char: index
            for index, char in enumerate(sorted(list(char_set)), start=len(self.char_dict))
        })
        self.num_classes = len(self.char_dict)

        self.common_feature_list = []
        self.unique_feature_list = []
        logger.info("Init ML validator with %s provider; config:'%s' md5:%s model:'%s' md5:%s", providers,
                    ml_config_path, md5_config, ml_model_path, md5_model)
        logger.debug("ML validator details: %s", model_config)
        for feature_definition in model_config["features"]:
            feature_class = feature_definition["type"]
            kwargs = feature_definition.get("kwargs", {})
            feature_constructor = getattr(features, feature_class, None)
            if feature_constructor is None:
                raise ValueError(f'Error while parsing model details. Cannot create feature "{feature_class}"')
            try:
                feature = feature_constructor(**kwargs)
            except TypeError:
                raise TypeError(f'Error while parsing model details. Cannot create feature "{feature_class}"'
                                f' with kwargs "{kwargs}"')
            if feature_definition["type"] in ["RuleName"]:
                self.unique_feature_list.append(feature)
            else:
                self.common_feature_list.append(feature)

    def encode(self, text: str, limit: int) -> np.ndarray:
        """Encodes prepared text to array"""
        result_array: np.ndarray = np.zeros(shape=(limit, self.num_classes), dtype=np.float32)
        if text is None:
            return result_array
        for i, c in enumerate(text):
            if i >= limit:
                break
            if c in self.char_dict:
                result_array[i, self.char_dict[c]] = 1.0
            else:
                result_array[i, self.char_dict[MlValidator.FAKE_CHAR]] = 1.0
        return result_array

    def encode_line(self, text: str, position: int):
        """Encodes line with balancing for position"""
        offset = len(text) - len(text.lstrip())
        pos = position - offset
        stripped = text.strip()
        if MlValidator.MAX_LEN < len(stripped):
            stripped = Util.subtext(stripped, pos, ML_HUNK)
        return self.encode(stripped, MlValidator.MAX_LEN)

    def encode_value(self, text: str) -> np.ndarray:
        """Encodes line with balancing for position"""
        stripped = text.strip()
        return self.encode(stripped[:ML_HUNK], ML_HUNK)

    def _call_model(self, line_input: np.ndarray, variable_input: np.ndarray, value_input: np.ndarray,
                    feature_input: np.ndarray) -> np.ndarray:
        input_feed: Dict[str, np.ndarray] = {
            "line_input": line_input.astype(np.float32),
            "variable_input": variable_input.astype(np.float32),
            "value_input": value_input.astype(np.float32),
            "feature_input": feature_input.astype(np.float32),
        }
        result = self.model_session.run(output_names=None, input_feed=input_feed)
        if result and isinstance(result[0], np.ndarray):
            return result[0]
        raise RuntimeError(f"Unexpected type {type(result[0])}")

    def extract_common_features(self, candidates: List[Candidate]) -> np.ndarray:
        """Extract features that are guaranteed to be the same for all candidates on the same line with same value."""
        feature_array: np.ndarray = np.array([], dtype=np.float32)
        # Extract features from credential candidate
        default_candidate = candidates[0]
        for feature in self.common_feature_list:
            new_feature = feature([default_candidate])[0]
            if not isinstance(new_feature, np.ndarray):
                new_feature = np.array([new_feature])
            feature_array = np.append(feature_array, new_feature)
        return feature_array

    def extract_unique_features(self, candidates: List[Candidate]) -> np.ndarray:
        """Extract features that can be different between candidates. Join them with or operator."""
        feature_array: np.ndarray = np.array([], dtype=np.int8)
        default_candidate = candidates[0]
        for feature in self.unique_feature_list:
            new_feature = feature([default_candidate])[0]
            if not isinstance(new_feature, np.ndarray):
                new_feature = np.array([new_feature])
            feature_array = np.append(feature_array, new_feature)
        for candidate in candidates[1:]:
            for feature in self.unique_feature_list:
                new_feature = feature([candidate])[0]
                if not isinstance(new_feature, np.ndarray):
                    new_feature = np.array([new_feature])
                feature_array = feature_array | new_feature
        return feature_array

    def get_group_features(self, candidates: List[Candidate]) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        `np.newaxis` used to add new dimension if front, so input will be treated as a batch
        """
        # all candidates are from the same line
        default_candidate = candidates[0]
        line_input = self.encode_line(default_candidate.line_data_list[0].line,
                                      default_candidate.line_data_list[0].value_start)[np.newaxis]
        variable = ""
        value = ""
        for candidate in candidates:
            if not variable and candidate.line_data_list[0].variable:
                variable = candidate.line_data_list[0].variable
            if not value and candidate.line_data_list[0].value:
                value = candidate.line_data_list[0].value
            if variable and value:
                break
        variable_input = self.encode_value(variable)[np.newaxis]
        value_input = self.encode_value(value)[np.newaxis]
        feature_array = self.extract_features(candidates)
        return line_input, variable_input, value_input, feature_array

    def extract_features(self, candidates: List[Candidate]) -> np.ndarray:
        """extracts common and unique features from list of candidates"""
        common_features = self.extract_common_features(candidates)
        unique_features = self.extract_unique_features(candidates)
        feature_hstack = np.hstack([common_features, unique_features])
        feature_array = np.array([feature_hstack])
        return feature_array

    def _batch_call_model(self, line_input_list, variable_input_list, value_input_list, features_list) -> np.ndarray:
        """auxiliary method to invoke twice"""
        line_inputs_vstack = np.vstack(line_input_list)
        variable_inputs_vstack = np.vstack(variable_input_list)
        value_inputs_vstack = np.vstack(value_input_list)
        feature_array_vstack = np.vstack(features_list)
        result_call = self._call_model(line_inputs_vstack, variable_inputs_vstack, value_inputs_vstack,
                                       feature_array_vstack)
        result = result_call[:, 0]
        return result

    def validate_groups(self, group_list: List[Tuple[CandidateKey, List[Candidate]]],
                        batch_size: int) -> Tuple[np.ndarray, np.ndarray]:
        """Use ml model on list of candidate groups.

        Args:
            group_list: List of tuples (value, group)
            batch_size: ML model batch

        Return:
            Boolean numpy array with decision based on the threshold,
            and numpy array with probability predicted by the model

        """
        line_input_list = []
        variable_input_list = []
        value_input_list = []
        features_list = []
        probability: np.ndarray = np.zeros(len(group_list), dtype=np.float32)
        head = tail = 0
        for group_key, candidates in group_list:
            line_input, variable_input, value_input, feature_array = self.get_group_features(candidates)
            line_input_list.append(line_input)
            variable_input_list.append(variable_input)
            value_input_list.append(value_input)
            features_list.append(feature_array)
            tail += 1
            if 0 == tail % batch_size:
                # use the approach to reduce memory consumption for huge candidates list
                probability[head:tail] = self._batch_call_model(line_input_list, variable_input_list, value_input_list,
                                                                features_list)
                head = tail
                line_input_list.clear()
                variable_input_list.clear()
                value_input_list.clear()
                features_list.clear()
        if head != tail:
            probability[head:tail] = self._batch_call_model(line_input_list, variable_input_list, value_input_list,
                                                            features_list)
        is_cred = probability > self.threshold
        if logger.isEnabledFor(logging.DEBUG):
            for i in range(len(is_cred)):
                logger.debug("ML decision: %s with prediction: %s for value: %s", is_cred[i], probability[i],
                             group_list[i][0])
        # apply cast to float to avoid json export issue
        return is_cred, probability.astype(float)



from functools import cached_property
from typing import Optional

from credsweeper.common.constants import Chars, ENTROPY_LIMIT_BASE64, ENTROPY_LIMIT_BASE3x
from credsweeper.utils import Util


class EntropyValidator:
    """Verifies data entropy with base64, base36 and base16(hex)"""
    CHARS_LIMIT_MAP = {
        Chars.BASE64STDPAD_CHARS: ENTROPY_LIMIT_BASE64,
        Chars.BASE36_CHARS: ENTROPY_LIMIT_BASE3x,
        Chars.HEX_CHARS: ENTROPY_LIMIT_BASE3x
    }

    def __init__(self, data: str, iterator: Optional[Chars] = None):
        self.__valid: Optional[bool] = None
        self.__entropy: Optional[float] = None
        self.__iterator: Optional[str] = None
        if isinstance(data, str):
            if isinstance(iterator, Chars):
                self.__entropy = Util.get_shannon_entropy(data, iterator.value)
                if _limit := self.CHARS_LIMIT_MAP.get(iterator):
                    self.__valid = _limit < self.__entropy
                    self.__iterator = iterator.name
            else:
                for _iterator, _limit in self.CHARS_LIMIT_MAP.items():
                    entropy = Util.get_shannon_entropy(data, _iterator.value)
                    if _limit < entropy:
                        self.__entropy = entropy
                        self.__iterator = _iterator.name
                        self.__valid = True
                        break
                    else:
                        # keep maximal entropy value
                        if self.__entropy:
                            if self.__entropy < entropy:
                                self.__entropy = entropy
                                self.__iterator = _iterator.name
                        else:
                            self.__entropy = entropy
                            self.__iterator = _iterator.name
                else:
                    self.__valid = False

    @cached_property
    def valid(self) -> Optional[bool]:
        """Shows whether validation was successful"""
        return self.__valid

    @cached_property
    def entropy(self) -> Optional[float]:
        """Value success entropy or maximal value"""
        return self.__entropy

    @cached_property
    def iterator(self) -> Optional[str]:
        """Which iterator was used for the entropy"""
        return self.__iterator

    def __repr__(self) -> str:
        if isinstance(self.entropy, float):
            return f"{self.iterator} {self.entropy:.6f} {self.valid}"
        else:
            return f"{self.iterator} {self.entropy} {self.valid}"

    def __str__(self) -> str:
        return self.__repr__()

    def to_dict(self) -> dict:
        """Representation to dictionary"""
        return {"iterator": self.iterator, "entropy": self.entropy, "valid": self.valid}



from credsweeper.common import static_keyword_checklist
from credsweeper.common.constants import Chars
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter, ValueEntropyBase64Check
from credsweeper.utils import Util


class ValueFilePathCheck(Filter):
    """Check that candidate value is a path or not.

    Check if a value contains either '/' or ':\' separators (but not both)
    and do not have any special characters ( !$@`&*()+)
    """
    base64stdpad_possible_set = set(Chars.BASE64STDPAD_CHARS.value)
    unusual_windows_symbols_in_path = "\t\n\r!$@`&*(){}<>+=;,~^"
    unusual_linux_symbols_in_path = "\t\n\r!@`&*<>+=;,~^:\\"

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        value = line_data.value
        contains_unix_separator = '/' in value
        if contains_unix_separator:
            if ("://" in value  #
                    or value.startswith("~/")  #
                    or value.startswith("./")  #
                    or "../" in value  #
                    or "/.." in value  #
                    or value.startswith("//") and ':' == line_data.separator):
                # common case for url definition or aliases
                # or _keyword_://example.com where : is the separator
                return static_keyword_checklist.check_morphemes(value.lower(), 1)
            # base64 encoded data might look like linux path
            min_entropy = ValueEntropyBase64Check.get_min_data_entropy(len(value))
            # get minimal entropy to compare with shannon entropy of found value
            # min_entropy == 0 means that the value cannot be checked with the entropy due high variance
            for i in value:
                if i not in self.base64stdpad_possible_set:
                    # value contains wrong BASE64STDPAD_CHARS symbols like -_
                    break
            else:
                # all symbols are from base64 alphabet
                entropy = Util.get_shannon_entropy(value, Chars.BASE64STDPAD_CHARS.value)
                if 0 == min_entropy or min_entropy > entropy:
                    contains_unix_separator = 1 < value.count('/')
                else:
                    # high entropy means base64 encoded data
                    contains_unix_separator = False

            # low shannon entropy points that the value maybe not a high randomized value in base64
        contains_windows_separator = ':\\' in value
        if contains_unix_separator or contains_windows_separator:
            unusual_symbols_in_path = self.unusual_linux_symbols_in_path if contains_unix_separator \
                else self.unusual_windows_symbols_in_path
            for i in unusual_symbols_in_path:
                if i in value:
                    # the symbols which not passed in a path usually
                    break
            else:
                if contains_unix_separator ^ contains_windows_separator:
                    return static_keyword_checklist.check_morphemes(value.lower(), 1)
        return False



import logging
import re
from abc import ABC, abstractmethod
from typing import List

from credsweeper.common.constants import RuleType, MIN_DATA_LEN
from credsweeper.config import Config
from credsweeper.credentials import Candidate, LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.rules import Rule

logger = logging.getLogger(__name__)


class ScanType(ABC):
    """Base class for all Scanners.

    Scanner allow to check if regex pattern defined in a rule is present in a line.

    """

    @classmethod
    @abstractmethod
    def run(cls, config: Config, rule: Rule, target: AnalysisTarget) -> List[Candidate]:
        """Check if regex pattern defined in a rule is present in a line.

        Args:
            config: user configs
            rule: Rule object to check current line
            target: Analysis target

        Return:
            List of Candidate objects if pattern defined in a rule is present in a line
            and filters defined in rule do not remove current line. Empty list - otherwise

        """
        raise NotImplementedError()

    @classmethod
    def filtering(cls, config: Config, target: AnalysisTarget, line_data: LineData, filters: List[Filter]) -> bool:
        """Check if line data should be removed based on filters.

        If `use_filters` option is false, always return False

        Args:
            config: dict of credsweeper configuration
            target: AnalysisTarget from which `line_data` was obtained
            line_data: Line data to check with `filters`
            filters: Filters to use

        Return:
            boolean: True if line_data should be removed. False otherwise.
            If `use_filters` option is false, always return False

        """
        if not line_data.value:
            logger.debug("Filtered line with empty value in file: %s:%d  in line: %s value: '%s'", line_data.path,
                         line_data.line_num, line_data.line, line_data.value)
            return True
        for filter_ in filters:
            if filter_.run(line_data, target):
                logger.debug("Filtered line with filter: %s in file: %s:%d  in line: %s value: %s",
                             filter_.__class__.__name__, line_data.path, line_data.line_num, line_data.line,
                             line_data.value)
                return True
        return False

    @classmethod
    def get_line_data_list(
            cls,  #
            config: Config,  #
            target: AnalysisTarget,  #
            pattern: re.Pattern,  #
            filters: List[Filter]) -> List[LineData]:
        """Check if regex pattern is present in line, and line should not be removed by filters.

        Args:
            config: dict of credsweeper configuration
            target: AnalysisTarget with all necessary data
            pattern: Compiled regex object to be searched in line
            filters: Filters to use

        Return:
            List of LineData objects if pattern a line and filters do not remove current line. Empty otherwise

        """
        line_data_list: List[LineData] = []
        # start - end positions for continuously searching for overlapping pattern
        offsets = [(0, target.line_len)]

        while offsets:
            offset_start, offset_end = offsets.pop()
            bypass_start = bypass_end = None
            for _match in pattern.finditer(target.line, pos=offset_start, endpos=offset_end):

                logger.debug("Valid line for pattern: %s in file: %s:%d in line: %s", pattern.pattern, target.file_path,
                             target.line_num, target.line)
                line_data = LineData(config, target.line, target.line_pos, target.line_num, target.file_path,
                                     target.file_type, target.info, pattern, _match)
                if bypass_start and bypass_end:
                    if 0 < line_data.variable_start:
                        bypass_end = line_data.variable_start
                    elif 0 < line_data.value_start:
                        bypass_end = line_data.value_start
                    if bypass_start < bypass_end and bypass_end - bypass_start > MIN_DATA_LEN:
                        offsets.append((bypass_start, bypass_end))
                    bypass_start = bypass_end = None
                elif MIN_DATA_LEN < line_data.value_end < _match.end() \
                        and MIN_DATA_LEN < _match.end() - line_data.value_end:
                    # add bypass for valuable sanitized value
                    bypass_start = line_data.value_end
                    bypass_end = offset_end

                if config.use_filters and cls.filtering(config, target, line_data, filters):
                    if line_data.variable and 0 <= line_data.variable_start < line_data.variable_end:
                        # may be next matched item will be not filtered - let search it after variable
                        bypass_start = line_data.variable_end
                        bypass_end = offset_end
                    elif line_data.value and 0 <= line_data.value_start < line_data.value_end:
                        # may be next matched item will be not filtered - let search it after variable
                        bypass_start = line_data.value_end
                        bypass_end = offset_end
                    continue

                if target.offset is not None:
                    # the target line is a chunk of long line - offsets have to be corrected
                    if 0 <= line_data.variable_start:
                        line_data.variable_start += target.offset
                    if 0 <= line_data.variable_end:
                        line_data.variable_end += target.offset
                    if 0 <= line_data.separator_start:
                        line_data.separator_start += target.offset
                    if 0 <= line_data.separator_end:
                        line_data.separator_end += target.offset
                    # value positions are mandatory
                    line_data.value_start += target.offset
                    line_data.value_end += target.offset
                    # get the original line
                    line_data.line = target.lines[target.line_pos]

                line_data_list.append(line_data)
            if bypass_start and bypass_end:
                offsets.append((bypass_start, bypass_end))

        return line_data_list

    @classmethod
    def _get_candidates(cls, config: Config, rule: Rule, target: AnalysisTarget) -> List[Candidate]:
        """Returns Candidate objects list.

        Args:
            config: user configs
            rule: Rule object to check current line
            target: Target for analysis

        Return:
            List of Candidate objects if pattern defined in a rule is present in a line
            and filters defined in rule do not remove current line. Empty list - otherwise

        """
        candidates: List[Candidate] = []
        if config.exclude_lines and target.line_strip in config.exclude_lines:
            return candidates

        if line_data_list := cls.get_line_data_list(config=config,
                                                    target=target,
                                                    pattern=rule.patterns[0],
                                                    filters=rule.filters):
            for line_data in line_data_list:
                if config.exclude_values and line_data.value.strip() in config.exclude_values:
                    continue
                candidate = Candidate(line_data_list=[line_data],
                                      patterns=rule.patterns,
                                      rule_name=rule.rule_name,
                                      severity=rule.severity,
                                      config=config,
                                      use_ml=rule.use_ml,
                                      confidence=rule.confidence)
                # single pattern with multiple values means all the patterns must matched in target
                if 1 < len(rule.patterns) and rule.rule_type in (RuleType.PATTERN, RuleType.KEYWORD):
                    # additional check whether all patterns match
                    if not cls._aux_scan(config, rule, target, candidate):
                        # cannot find secondary values for the candidate
                        continue
                candidates.append(candidate)
        return candidates

    @classmethod
    def _aux_scan(cls, config: Config, rule: Rule, target: AnalysisTarget, candidate: Candidate) -> bool:
        """check for all secondary patterns"""
        for pattern in rule.patterns[1:]:
            line_data_list = cls.get_line_data_list(config=config, target=target, pattern=pattern, filters=rule.filters)
            pattern_matched = False

            for line_data in line_data_list:
                # standard filtering of values from config
                if config.exclude_values and line_data.value.strip() in config.exclude_values:
                    continue
                candidate.line_data_list.append(line_data)
                pattern_matched = True
            if not pattern_matched:
                return False

        # all secondary patterns were matched and candidate is filled with the values
        return True



import contextlib
import json

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueJsonWebTokenCheck(Filter):
    """
    Check that candidate is JWT which starts usually from 'eyJ'
    registered keys are checked to be in the JWT parts
    https://www.iana.org/assignments/jose/jose.xhtml
    """
    header_keys = {
        "alg", "jku", "jwk", "kid", "x5u", "x5c", "x5t", "x5t#S256", "typ", "cty", "crit", "alg", "enc", "zip", "jku",
        "jwk", "kid", "x5u", "x5c", "x5t", "x5t#S256", "typ", "cty", "crit", "epk", "apu", "apv", "iv", "tag", "p2s",
        "p2c", "iss", "sub", "aud", "b64", "ppt", "url", "nonce", "svt"
    }
    payload_keys = {
        "iss", "sub", "aud", "exp", "nbf", "iat", "jti", "kty", "use", "key_ops", "alg", "enc", "zip", "jku", "jwk",
        "kid", "x5u", "x5c", "x5t", "x5t#S256", "crv", "x", "y", "d", "n", "e", "d", "p", "q", "dp", "dq", "qi", "oth",
        "k", "crv", "d", "x", "ext", "crit", "keys", "id", "role", "token", "secret", "password", "nonce"
    }

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        header_check = False
        payload_check = False
        signature_check = False
        with contextlib.suppress(Exception):
            jwt_parts = line_data.value.split('.')
            for part in jwt_parts:
                data = Util.decode_base64(part, padding_safe=True, urlsafe_detect=True)
                if part.startswith("eyJ"):
                    # open part - just base64 encoded
                    json_keys = json.loads(data).keys()
                    # header will be checked first
                    if not header_check:
                        header_check = bool(ValueJsonWebTokenCheck.header_keys.intersection(json_keys))
                    # payload follows the header
                    elif not payload_check:
                        payload_check = bool(ValueJsonWebTokenCheck.payload_keys.intersection(json_keys))
                        # any other payloads are allowed
                elif header_check and payload_check and not signature_check:
                    # signature check or skip encrypted part
                    signature_check = not Util.is_ascii_entropy_validate(data)
                else:
                    break
        if header_check and payload_check and signature_check:
            return False
        else:
            return True



from credsweeper.common.constants import GroupType
from credsweeper.config import Config
from credsweeper.filters import ValueDictionaryValueLengthCheck, LineGitBinaryCheck
from credsweeper.filters import ValueSplitKeywordCheck
from credsweeper.filters.group import Group
from credsweeper.filters.line_uue_part_check import LineUUEPartCheck


class PasswordKeyword(Group):
    """PasswordKeyword"""

    def __init__(self, config: Config) -> None:
        super().__init__(config, GroupType.KEYWORD)
        self.filters.extend(
            [ValueDictionaryValueLengthCheck(),
             ValueSplitKeywordCheck(),
             LineGitBinaryCheck(),
             LineUUEPartCheck()])



import logging
from typing import List

from credsweeper.common.constants import RuleType
from credsweeper.config import Config
from credsweeper.credentials import Candidate
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.rules import Rule
from credsweeper.scanner.scan_type import ScanType
from credsweeper.utils.pem_key_detector import PemKeyDetector

logger = logging.getLogger(__name__)


class PemKeyPattern(ScanType):
    """Scanner detects single PEM private key in target from current line"""

    @classmethod
    def run(cls, config: Config, rule: Rule, target: AnalysisTarget) -> List[Candidate]:
        """Check if target is a PEM key

        Args:
            config: user configs
            rule: Rule object to check current line. Should be a pem-pattern rule
            target: Analysis target

        Return:
            List of Candidate objects if pattern defined in a rule is present in a line
            and filters defined in rule do not remove current line. Empty list - otherwise

        """
        assert rule.rule_type == RuleType.PEM_KEY, \
            "Rules provided to PemKeyPattern.run should have pattern_type equal to PEM_KEY_PATTERN"
        if candidates := cls._get_candidates(config, rule, target):
            candidate = candidates[0]
            if pem_lines := PemKeyDetector.detect_pem_key(config, target):
                candidate.line_data_list = pem_lines
                return [candidate]

        return []



from credsweeper.ml_model.features.entropy_evaluation import EntropyEvaluation
from credsweeper.ml_model.features.file_extension import FileExtension
from credsweeper.ml_model.features.has_html_tag import HasHtmlTag
from credsweeper.ml_model.features.is_secret_numeric import IsSecretNumeric
from credsweeper.ml_model.features.length_of_attribute import LengthOfAttribute
from credsweeper.ml_model.features.morpheme_dense import MorphemeDense
from credsweeper.ml_model.features.rule_name import RuleName
from credsweeper.ml_model.features.search_in_attribute import SearchInAttribute
from credsweeper.ml_model.features.word_in_line import WordInLine
from credsweeper.ml_model.features.word_in_path import WordInPath
from credsweeper.ml_model.features.word_in_value import WordInValue
from credsweeper.ml_model.features.word_in_variable import WordInVariable



import binascii
import contextlib

import base62

from credsweeper.common.constants import ASCII
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueGitHubCheck(Filter):
    """GitHub Classic Token validation"""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, when need to filter candidate and False if left

        """
        # https://github.blog/2021-04-05-behind-githubs-new-authentication-token-formats/
        with contextlib.suppress(Exception):
            if line_data.value.startswith("gh") and '_' == line_data.value[3]:
                token = line_data.value[4:-6]
                data = token.encode(ASCII, errors="strict")
                crc32sum = binascii.crc32(data)
                base62_crc32 = line_data.value[-6:]
                sign_b = base62.decodebytes(base62_crc32)
                crc32sign = int.from_bytes(sign_b, "big")
                if crc32sign == crc32sum:
                    return False
        return True



from functools import cached_property
from typing import List, Optional, Generator

from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.content_provider import ContentProvider


class StringContentProvider(ContentProvider):
    """Provider performs scan simple text lines"""

    def __init__(
            self,  #
            lines: List[str],  #
            line_numbers: Optional[List[int]] = None,  #
            file_path: Optional[str] = None,  #
            file_type: Optional[str] = None,  #
            info: Optional[str] = None) -> None:
        """
        Parameters:
            lines: text lines to be processed
            line_numbers: matched line numbers for lines if the order is not natural.
                Otherwise, it will be filled with natural order from 1.

        """
        super().__init__(file_path=file_path, file_type=file_type, info=info)
        self.__lines = lines
        # fill line numbers only when amounts are equal
        if line_numbers is None or len(lines) != len(line_numbers):
            self.__line_numbers = None
        else:
            self.__line_numbers = line_numbers

    @cached_property
    def data(self) -> bytes:
        """data getter for StringContentProvider"""
        raise NotImplementedError(__name__)

    def free(self) -> None:
        """free data after scan to reduce memory usage"""
        self.__lines = []
        if hasattr(self, "lines"):
            delattr(self, "lines")
        self.__line_numbers = []
        if hasattr(self, "line_numbers"):
            delattr(self, "line_numbers")

    @cached_property
    def lines(self) -> List[str]:
        """line_numbers RO getter for StringContentProvider"""
        return self.__lines

    @cached_property
    def line_numbers(self) -> List[int]:
        """line_numbers RO getter for StringContentProvider"""
        if self.__line_numbers is None or len(self.__lines) != len(self.__line_numbers):
            self.__line_numbers = list(range(1, 1 + len(self.__lines))) if self.__lines else []
        return self.__line_numbers

    def yield_analysis_target(self, min_len: int) -> Generator[AnalysisTarget, None, None]:
        """Return lines to scan.

        Args:
            min_len: minimal line length to scan

        Return:
            list of analysis targets based on every row in file

        """
        return self.lines_to_targets(min_len, self.lines, self.line_numbers)



import copy
import re
from json.encoder import py_encode_basestring_ascii
from typing import Any, Dict, List, Optional

from credsweeper.common.constants import Severity, Confidence
from credsweeper.config import Config
from credsweeper.credentials.line_data import LineData


class Candidate:
    """Candidates that can be credentials.

    Class contains list of LineData, some attributes from Rule object, and config

    Parameters:
        line_data_list: List of LineData
        patterns: Regular expressions that can be used for detection
        rule_name: Name of Rule
        severity: critical/high/medium/low
        confidence: strong/moderate/weak
        config: user configs
        use_ml: Whether the candidate should be validated with ML. If not - ml_probability is set None
    """

    DUMMY_PATTERN = re.compile(r"^")

    def __init__(self,
                 line_data_list: List[LineData],
                 patterns: List[re.Pattern],
                 rule_name: str,
                 severity: Severity,
                 config: Optional[Config] = None,
                 use_ml: bool = False,
                 confidence: Confidence = Confidence.MODERATE) -> None:
        self.line_data_list = line_data_list
        self.patterns = patterns
        self.rule_name = rule_name
        self.severity = severity
        self.config = config
        self.use_ml = use_ml
        self.confidence = confidence
        # None - ML is not applicable or not processed yet; float - the ml decision above ml_threshold
        # Note: -1.0 is possible too for some activation functions in ml model, so let avoid negative values
        self.ml_probability: Optional[float] = None

    def compare(self, other: 'Candidate') -> bool:
        """Comparison method - checks only result of final cred"""
        if self.rule_name == other.rule_name \
                and self.severity == other.severity \
                and self.confidence == other.confidence \
                and self.use_ml == other.use_ml \
                and self.ml_probability == other.ml_probability \
                and len(self.line_data_list) == len(other.line_data_list):
            for i, j in zip(self.line_data_list, other.line_data_list):
                if i.compare(j):
                    continue
                else:
                    break
            else:
                # all line_data are equal
                return True
        return False

    @staticmethod
    def _encode(value: Any) -> Any:
        """Encode value to the base string ascii

        Args:
            value: Any type of value to be encoded
        """
        if isinstance(value, str):
            return py_encode_basestring_ascii(value)
        else:
            return value

    def to_str(self, subtext: bool = False, hashed: bool = False) -> str:
        """Represent candidate with subtext or|and hashed values"""
        return f"rule: {self.rule_name}" \
               f" | severity: {self.severity.value}" \
               f" | confidence: {self.confidence.value}" \
               f" | ml_probability: {self.ml_probability}" \
               f" | line_data_list: [{', '.join([x.to_str(subtext, hashed) for x in self.line_data_list])}]"

    def __str__(self):
        return self.to_str()

    def __repr__(self):
        return self.to_str(subtext=True)

    def to_json(self, hashed: bool, subtext: bool) -> Dict:
        """Convert credential candidate object to dictionary.

        Return:
            Dictionary object generated from current credential candidate

        """
        full_output = {
            "patterns": [pattern.pattern for pattern in self.patterns],
            "rule": self.rule_name,
            "severity": self.severity.value,
            "confidence": self.confidence.value,
            "use_ml": self.use_ml,
            "ml_probability": self.ml_probability,
            # put the array to end to make json more readable
            "line_data_list": [line_data.to_json(hashed, subtext) for line_data in self.line_data_list],
        }
        if self.config is not None:
            reported_output = {k: v for k, v in full_output.items() if k in self.config.candidate_output}
        else:
            reported_output = full_output
        return reported_output

    def to_dict_list(self, hashed: bool, subtext: bool) -> List[dict]:
        """Convert credential candidate object to List[dict].

        Return:
            List[dict] object generated from current credential candidate

        """
        reported_output = []
        json_output = self.to_json(hashed, subtext)
        refined_data = copy.deepcopy(json_output)
        del refined_data["line_data_list"]
        for line_data in json_output["line_data_list"]:
            line_data.update(refined_data)
            for key in line_data.keys():
                line_data[key] = self._encode(line_data[key])
            reported_output.append(line_data)
        return reported_output

    @classmethod
    def get_dummy_candidate(cls, config: Config, file_path: str, file_type: str, info: str, rule_name: str):
        """Create dummy instance to use in searching file by extension"""
        return cls(  #
            line_data_list=[LineData(config, '', -1, 0, file_path, file_type, info, cls.DUMMY_PATTERN)],
            patterns=[cls.DUMMY_PATTERN],  #
            rule_name=rule_name,  #
            severity=Severity.INFO,  #
            config=config,  #
            confidence=Confidence.WEAK)



from credsweeper.common import static_keyword_checklist
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueDictionaryKeywordCheck(Filter):
    """Check that no word from dictionary present in the candidate value."""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        line_data_value_lower = line_data.value.lower()
        for keyword in static_keyword_checklist.keyword_list:
            if keyword in line_data_value_lower:
                line_data_value_lower = line_data_value_lower.replace(keyword, '\x7F' * len(keyword))
                ratio = line_data_value_lower.count('\x7F') / len(line_data_value_lower)
                if 0.33 < ratio:
                    return True
        return False



from abc import abstractmethod, ABC

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget


class Filter(ABC):
    """Base class for all filters that operates on 'line_data' objects."""

    @abstractmethod
    def __init__(self, config: Config, *args):
        raise NotImplementedError()

    @abstractmethod
    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        raise NotImplementedError()



import contextlib

from credsweeper.common.constants import ASCII, PEM_BEGIN_PATTERN
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util
from credsweeper.utils.pem_key_detector import PemKeyDetector


class ValueBase64EncodedPem(Filter):
    """Check that candidate contains base64 encoded pem private key"""

    def __init__(self, config: Config = None) -> None:
        self.config = config

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """

        with contextlib.suppress(Exception):
            text = Util.decode_base64(line_data.value, padding_safe=True, urlsafe_detect=True)
            lines = text.decode(ASCII).splitlines()
            lines_pos = [x for x in range(len(lines))]
            for line_pos, line in zip(lines_pos, lines):
                if PEM_BEGIN_PATTERN in line:
                    new_target = AnalysisTarget(line_pos, lines, lines_pos, target.descriptor)
                    if PemKeyDetector.detect_pem_key(self.config, new_target):
                        # obtained candidates are not used because not match text
                        return False
        return True



from typing import List, Any

import numpy as np

from credsweeper.credentials import Candidate
from credsweeper.ml_model.features.word_in import WordIn


class FileExtension(WordIn):
    """Categorical feature of file type.

    Parameters:
        extensions: extension labels

    """

    def __init__(self, extensions: List[str]) -> None:
        super().__init__(extensions)

    def __call__(self, candidates: List[Candidate]) -> np.ndarray:
        extension_set = set([candidate.line_data_list[0].file_type.lower() for candidate in candidates])
        return self.word_in_set(extension_set)

    def extract(self, candidate: Candidate) -> Any:
        raise NotImplementedError



import contextlib
from abc import abstractmethod
from typing import Tuple

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils.hop_stat import HopStat


class ValueTokenBaseCheck(Filter):
    """Check that candidate have good randomization"""

    MUL_DICT = {
        8: 2.61619746,
        10: 2.48685659,
        15: 2.34025271,
        16: 2.32370290,
        20: 2.27614996,
        24: 2.24609586,
        25: 2.24023515,
        32: 2.21025277,
        40: 2.18961571,
        50: 2.17355282,
        64: 2.15981241,
    }

    def __init__(self, config: Config = None) -> None:
        self.__hop_stat = HopStat()

    @staticmethod
    @abstractmethod
    def get_stat_range(size: int) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Returns minimal strength. Precalculated data is applied for speedup"""
        raise NotImplementedError

    @staticmethod
    def get_ppf(n: int) -> float:
        """
        from scipy.stats import t
        print('\n'.join(f'{n}: {t.ppf(0.9827, n-1):.8f},' for n in [8,10,15,16,20,24,25,32,40,50,64]))
        """
        return ValueTokenBaseCheck.MUL_DICT[n]

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        with contextlib.suppress(Exception):
            hop, dev = self.__hop_stat.stat(line_data.value)
            (min_hop, max_hop), (min_dev, max_dev) = self.get_stat_range(len(line_data.value))
            if not (min_hop <= hop <= max_hop and min_dev <= dev <= max_dev):
                return True
        return False



import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueNumberCheck(Filter):
    """Check value if it a value in hex or decimal representation"""

    HEX_VALUE_REGEX = re.compile("^(0x)?[0-9a-f]{1,128}[ul]{0,3}$")
    DEC_VALUE_REGEX = re.compile("^-?[0-9]{1,20}[ul]{0,3}$")

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        value = line_data.value.lower()
        if 22 > len(value) and ValueNumberCheck.HEX_VALUE_REGEX.match(value):
            return True
        if ValueNumberCheck.DEC_VALUE_REGEX.match(value):
            return True
        return False



import logging
import re
from pathlib import Path
from typing import List, Type, Tuple, Union, Dict, Generator, Set

from credsweeper.app import APP_PATH
from credsweeper.common.constants import RuleType, MIN_VARIABLE_LENGTH, MIN_SEPARATOR_LENGTH, MIN_VALUE_LENGTH, \
    MAX_LINE_LENGTH, PEM_BEGIN_PATTERN
from credsweeper.config import Config
from credsweeper.credentials import Candidate
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.content_provider import ContentProvider
from credsweeper.rules import Rule
from credsweeper.scanner.scan_type import PemKeyPattern, ScanType, SinglePattern, MultiPattern
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class Scanner:
    """Advanced Credential Scanner base class.

    Parameters:
        rules: list of rule objects to check
        min_pattern_len: minimal length specified in all pattern rules
        min_keyword_len: minimal possible length for a string to be matched by any keyword rule
        min_len: Smallest between min_pattern_len and min_keyword_len
        TargetGroup: Type for List[Tuple[AnalysisTarget, str, int]]

    """

    TargetGroup = List[Tuple[AnalysisTarget, str, int]]

    def __init__(self, config: Config, rule_path: Union[None, str, Path]) -> None:
        self.config = config
        # init with MAX_LINE_LENGTH before _set_rules
        self.min_keyword_len = MAX_LINE_LENGTH
        self.min_pattern_len = MAX_LINE_LENGTH
        self.min_pem_key_len = MAX_LINE_LENGTH
        self.min_multi_len = MAX_LINE_LENGTH
        self.rules_scanners: List[Tuple[Rule, Type[ScanType]]] = []
        self._set_rules_scanners(rule_path)
        self.min_len = min(self.min_pattern_len, self.min_keyword_len, self.min_pem_key_len, self.min_multi_len,
                           MIN_VARIABLE_LENGTH + MIN_SEPARATOR_LENGTH + MIN_VALUE_LENGTH)
        self.__keyword_rules_required_substrings = self._get_required_substrings(RuleType.KEYWORD)

    def keywords_required_substrings_check(self, text: str) -> bool:
        """check whether `text` has any required substring for all keyword type rules"""
        return self._substring_check(self.__keyword_rules_required_substrings, text)

    def _get_required_substrings(self, rule_type: RuleType) -> Set[str]:
        """init set of required substrings for custom rule type"""
        required_substrings: Set[str] = set()
        for rule in (x[0] for x in self.rules_scanners if rule_type == x[0].rule_type):
            required_substrings.update(set(rule.required_substrings))
        return required_substrings

    @staticmethod
    def _substring_check(substrings: Set[str], text: str) -> bool:
        """checks whether `text` has any required substring. Set is used to reduce extra transformations"""
        for substring in substrings:
            if substring in text:
                return True
        return False

    def _set_rules_scanners(self, rule_path: Union[None, str, Path]) -> None:
        """Auxiliary method to fill rules, determine min_pattern_len and set scanners"""
        if rule_path is None:
            rule_path = APP_PATH / "rules" / "config.yaml"
        rule_templates = Util.yaml_load(rule_path)
        if rule_templates and isinstance(rule_templates, list):
            for rule_template in rule_templates:
                try:
                    rule = Rule(self.config, rule_template)
                except Exception as exc:
                    logger.error("Rule creation error%s", str(rule_template))
                    raise exc
                if not self._is_available(rule):
                    continue
                if 0 < rule.min_line_len:
                    if rule.rule_type == RuleType.KEYWORD:
                        self.min_keyword_len = min(self.min_keyword_len, rule.min_line_len)
                    elif rule.rule_type == RuleType.PATTERN:
                        self.min_pattern_len = min(self.min_pattern_len, rule.min_line_len)
                    elif rule.rule_type == RuleType.PEM_KEY:
                        self.min_pem_key_len = min(self.min_pem_key_len, rule.min_line_len)
                    elif rule.rule_type == RuleType.MULTI:
                        self.min_multi_len = min(self.min_multi_len, rule.min_line_len)
                    else:
                        logger.warning(f"Unknown rule type:{rule.rule_type}")
                self.rules_scanners.append((rule, self.get_scanner(rule)))
        else:
            raise RuntimeError(f"Wrong rules '{rule_templates}' were read from '{rule_path}'")

    def _is_available(self, rule: Rule) -> bool:
        """separate the method to reduce complexity"""
        if rule.severity < self.config.severity:
            return False
        if self.config.doc:
            if "doc" in rule.target:
                return True
        else:
            if "code" in rule.target:
                return True
        return False

    def yield_rule_scanner(
            self,  #
            line_len: int,  #
            matched_pattern: bool,  #
            matched_keyword: bool,  #
            matched_pem_key: bool,  #
            matched_multi: bool) -> Generator[Tuple[Rule, Type[ScanType]], None, None]:
        """returns generator for rules and according scanner"""
        for rule, scanner in self.rules_scanners:
            if line_len >= rule.min_line_len \
                    and (RuleType.PATTERN == rule.rule_type and matched_pattern
                         or RuleType.KEYWORD == rule.rule_type and matched_keyword
                         or RuleType.PEM_KEY == rule.rule_type and matched_pem_key
                         or RuleType.MULTI == rule.rule_type and matched_multi):
                yield rule, scanner

    def scan(self, provider: ContentProvider) -> List[Candidate]:
        """Run scanning of list of target lines from 'targets' with set of rule from 'self.rules'.

        Args:
            provider: objects with data to analyze: line, line number,
              filepath and all lines in file

        Return:
            list of all detected credential candidates in analyzed targets

        """
        credentials: List[Candidate] = []

        for target in provider.yield_analysis_target(self.min_len):
            # Trim string from outer spaces to make future `x in str` checks faster
            target_line_stripped = target.line_strip
            target_line_stripped_len = target.line_strip_len

            # "cache" - YAPF and pycharm formatters ...
            matched_keyword = \
                target_line_stripped_len >= self.min_keyword_len and (  #
                    '=' in target_line_stripped or ':' in target_line_stripped)  #
            matched_pem_key = \
                target_line_stripped_len >= self.min_pem_key_len \
                and PEM_BEGIN_PATTERN in target_line_stripped and "PRIVATE" in target_line_stripped
            matched_pattern = target_line_stripped_len >= self.min_pattern_len
            matched_multi = target_line_stripped_len >= self.min_multi_len

            if not (matched_keyword or matched_pem_key or matched_pattern or matched_multi):
                # target may be skipped only with length because not all rules have required_substrings
                logger.debug("Skip too short (%d) line %s:%d", target_line_stripped_len, target.file_path,
                             target.line_num)
                continue

            # use lower case for required substring
            target_line_stripped_lower = target.line_lower_strip
            # cached value to skip the same regex verifying
            matched_regex: Dict[re.Pattern, bool] = {}

            for rule, scanner in self.yield_rule_scanner(target_line_stripped_len, matched_pattern, matched_keyword,
                                                         matched_pem_key, matched_multi):
                if rule.has_required_substrings \
                        and not self._substring_check(rule.required_substrings, target_line_stripped_lower):
                    continue

                # common regex might be triggered for the same target
                if rule.required_regex:
                    if rule.required_regex in matched_regex:
                        regex_result = matched_regex[rule.required_regex]
                    else:
                        regex_result = bool(rule.required_regex.search(target_line_stripped))
                        matched_regex[rule.required_regex] = regex_result
                    if not regex_result:
                        continue

                if new_credentials := scanner.run(self.config, rule, target):
                    credentials.extend(new_credentials)
                    logger.debug("Credential for rule: %s in file: %s:%d in line: %s", rule.rule_name, target.file_path,
                                 target.line_num, target.line)
        return credentials

    @staticmethod
    def get_scanner(rule: Rule) -> Type[ScanType]:
        """Choose type of scanner base on rule affiliation.

        Args:
            rule: rule object used to scanning

        Return:
            depending on the rule type, returns the corresponding scanner class

        """
        if RuleType.PATTERN == rule.rule_type or RuleType.KEYWORD == rule.rule_type:
            return SinglePattern
        elif RuleType.MULTI == rule.rule_type:
            return MultiPattern
        elif RuleType.PEM_KEY == rule.rule_type:
            return PemKeyPattern
        raise ValueError(f"Unknown pattern_type in rule: {rule.rule_type}")



import re

from credsweeper.common.constants import DEFAULT_PATTERN_LEN
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValuePatternCheck(Filter):
    """Check if candidate value contain specific pattern.

    Similar to linguistic sequences of characters, random strings shouldn't contain math sequences of
    characters. Based on "How Bad Can It Git? Characterizing Secret Leakage in Public GitHub Repositories", details:
    https://www.ndss-symposium.org/ndss-paper/how-bad-can-it-git-characterizing-secret-leakage-in-public-github-repositories/
    PatternCheck checks the occurrence in "line_data.value" of three types of sequence:

    - N or more identical characters in sequence, example: "AAAA", "1111" ...
    - N or more increasing characters sequentially, example: "abcd", "1234" ...
    - N or more decreasing characters sequentially, example: "dcba", "4321" ...

    Default pattern LEN is 4
    """

    def __init__(self, config: Config = None, pattern_len: int = DEFAULT_PATTERN_LEN):
        """Create ValuePatternCheck with a specific pattern_len to check.

        Args:
            config: pattern len to use during check. DEFAULT_PATTERN_LEN by default

        """
        self.pattern_len = pattern_len
        # use non whitespace symbol pattern
        self.pattern = re.compile(fr"(\S)\1{{{str(self.pattern_len - 1)},}}")

    def equal_pattern_check(self, line_data_value: str) -> bool:
        """Check if candidate value contain 4 and more same chars or numbers sequences.

        Args:
            line_data_value: string variable, credential candidate value

        Return:
            True if contain and False if not

        """
        if self.pattern.findall(line_data_value):
            return True
        return False

    def ascending_pattern_check(self, line_data_value: str) -> bool:
        """Check if candidate value contain 4 and more ascending chars or numbers sequences.

        Arg:
            line_data_value: credential candidate value

        Return:
            True if contain and False if not

        """
        count = 1
        for key in range(len(line_data_value) - 1):
            if ord(line_data_value[key + 1]) - ord(line_data_value[key]) == 1:
                count += 1
            else:
                count = 1
                continue
            if count == self.pattern_len:
                return True
        return False

    def descending_pattern_check(self, line_data_value: str) -> bool:
        """Check if candidate value contain 4 and more descending chars or numbers sequences.

        Arg:
            line_data_value: string variable, credential candidate value

        Return:
            boolean variable. True if contain and False if not

        """
        count = 1
        for key in range(len(line_data_value) - 1):
            if ord(line_data_value[key]) - ord(line_data_value[key + 1]) == 1:
                count += 1
            else:
                count = 1
                continue
            if count == self.pattern_len:
                return True
        return False

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Arg:
            line_data: LineData object, credential candidate data
            target: multiline target from which line data was obtained

        Return:
            boolean variable. True, if need to filter candidate and False if left

        """
        if len(line_data.value) < self.pattern_len:
            return True

        if self.equal_pattern_check(line_data.value):
            return True

        if self.ascending_pattern_check(line_data.value):
            return True

        if self.descending_pattern_check(line_data.value):
            return True

        return False



import io
import logging
from pathlib import Path
from typing import List, Optional, Union, Tuple, Sequence

from credsweeper import DiffContentProvider
from credsweeper.config import Config
from credsweeper.file_handler.abstract_provider import AbstractProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.file_handler.text_content_provider import TextContentProvider

logger = logging.getLogger(__name__)


class FilesProvider(AbstractProvider):
    """Provider of plain os files to be analysed."""

    def __init__(self,
                 paths: Sequence[Union[str, Path, io.BytesIO, Tuple[Union[str, Path], io.BytesIO]]],
                 skip_ignored: Optional[bool] = None) -> None:
        """Initialize Files Text Provider for files from 'paths'.

        Args:
            paths: list of parent paths of files to scan
                   OR tuple of path (info purpose) and io.BytesIO (reads the data from current pos)
            skip_ignored: boolean variable, Checking the directory to the list
                          of ignored directories from the gitignore file

        """
        super().__init__(paths)
        self.skip_ignored = skip_ignored

    def get_scannable_files(self, config: Config) -> Sequence[Union[DiffContentProvider, TextContentProvider]]:
        """Get list of full text file object for analysis of files with parent paths from "paths".

        Args:
            config: dict of credsweeper configuration

        Return:
            preprocessed file objects for analysis

        """
        text_content_provider_list: List[Union[DiffContentProvider, TextContentProvider]] = []
        for path in self.paths:
            if isinstance(path, str) or isinstance(path, Path):
                new_files = FilePathExtractor.get_file_paths(config, path)
                if self.skip_ignored:
                    new_files = FilePathExtractor.apply_gitignore(new_files)
                for _file in new_files:
                    text_content_provider_list.append(TextContentProvider(_file))
            elif isinstance(path, io.BytesIO):
                text_content_provider_list.append(TextContentProvider((":memory:", path)))
            elif isinstance(path, tuple) \
                    and (isinstance(path[0], str) or isinstance(path[0], Path)) \
                    and isinstance(path[1], io.BytesIO):
                # suppose, all the files must be scanned
                text_content_provider_list.append(TextContentProvider(path))
            else:
                logger.error(f"Unknown path type: {path}")

        return text_content_provider_list



import logging
from multiprocessing import Manager
from typing import List, Dict, Tuple

from credsweeper.credentials import Candidate
from credsweeper.credentials.candidate_group_generator import CandidateGroupGenerator, CandidateKey

logger = logging.getLogger(__name__)


class CredentialManager:
    """The manager allows you to store, add and delete separate credit candidates.

    Parameters:
        candidates: list of credential candidates

    """

    def __init__(self) -> None:
        self.candidates: List[Candidate] = list(Manager().list())

    def get_credentials(self) -> List[Candidate]:
        """Get all credential candidates stored in the manager.

        Return:
            List with all Candidate objects stored in manager

        """
        return self.candidates

    def set_credentials(self, candidates: List[Candidate]) -> None:
        """Remove all current credentials candidates from the manager and add new credentials.

        Args:
            candidates: List with candidates to replace current candidates in the manager

        """
        self.candidates = candidates

    def add_credential(self, candidate: Candidate) -> None:
        """Add credential candidate to the manager.

        Args:
            candidate: credential candidate to be added

        """
        self.candidates.append(candidate)

    def remove_credential(self, candidate: Candidate) -> None:
        """Remove credential candidate from the manager.

        Args:
            candidate: credential candidate to be removed

        """
        self.candidates.remove(candidate)

    def purge_duplicates(self) -> int:
        """Purge duplicates candidates which may appear in overlaps during long line scan.

        Returns: number of removed duplicates
        """
        candidates_dict: Dict[Tuple[str, str, str, int, int, int, int, int, int, int], Candidate] = {}
        before = len(self.candidates)
        for i in self.candidates:
            ld = i.line_data_list[0]
            candidate_key = (
                i.rule_name,  #
                ld.path,  #
                ld.info,  #
                ld.line_pos,  #
                ld.variable_start,  #
                ld.variable_end,  #
                ld.separator_start,  #
                ld.separator_end,  #
                ld.value_start,  #
                ld.value_end)
            if candidate_key in candidates_dict:
                # check precisely - compare with the values
                candidate_dict = candidates_dict[candidate_key]
                if not candidate_dict.compare(i):
                    ld_ = candidate_dict.line_data_list[0]
                    logger.warning(f"check {ld_.variable, ld_.value} and {ld.variable, ld.value}")
            else:
                candidates_dict[candidate_key] = i
        self.candidates = list(candidates_dict.values())
        after = len(self.candidates)
        return before - after

    def group_credentials(self) -> CandidateGroupGenerator:
        """Join candidates that reference same secret value in the same line.

        Candidate can belong to two groups in the same time if it has more than one LineData object inside

        Return:
            Contain dictionary of [path, line_num, value] -> credential candidates list

        """
        groups = CandidateGroupGenerator()
        for credential_candidate in self.get_credentials():
            for line_data in credential_candidate.line_data_list[:1]:
                # Match by file path+line num+value. Value required so two different credentials still be
                #  processed independently
                candidate_key = CandidateKey(line_data)
                if candidate_key not in groups:
                    groups[candidate_key] = list()
                groups[candidate_key].append(credential_candidate)
        return groups



import io
import logging
from pathlib import Path
from typing import List, Union, Tuple, Sequence

from credsweeper import TextContentProvider
from credsweeper.common.constants import DiffRowType
from credsweeper.config import Config
from credsweeper.file_handler.abstract_provider import AbstractProvider
from credsweeper.file_handler.diff_content_provider import DiffContentProvider
from credsweeper.file_handler.file_path_extractor import FilePathExtractor
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class PatchesProvider(AbstractProvider):
    """Provide data from a list of `.patch` files.
    """

    def __init__(self, paths: Sequence[Union[str, Path, io.BytesIO, Tuple[Union[str, Path], io.BytesIO]]],
                 change_type: DiffRowType) -> None:
        """Initialize Files Patch Provider for patch files from 'paths'.

        Args:
            paths: file paths list to scan. All files should be in `.patch` format
            change_type: DiffRowType, type of analyses changes in patch (added or deleted)
              of ignored directories from the gitignore file

        """
        super().__init__(paths)
        self.change_type = change_type

    def load_patch_data(self, config: Config) -> List[List[str]]:
        """Loads data from patch"""
        raw_patches = []
        for file_path in self.paths:
            if FilePathExtractor.check_file_size(config, file_path):
                continue
            if isinstance(file_path, str) or isinstance(file_path, Path):
                raw_patches.append(Util.read_file(file_path))
            elif isinstance(file_path, io.BytesIO):
                the_patch = Util.decode_bytes(file_path.read())
                raw_patches.append(the_patch)
            else:
                logger.error(f"Unknown path type: {file_path}")

        return raw_patches

    def get_files_sequence(self,
                           raw_patches: List[List[str]]) -> Sequence[Union[DiffContentProvider, TextContentProvider]]:
        """Returns sequence of files"""
        files: List[Union[DiffContentProvider, TextContentProvider]] = []
        for raw_patch in raw_patches:
            files_data = Util.patch2files_diff(raw_patch, self.change_type)
            for file_path, file_diff in files_data.items():
                files.append(DiffContentProvider(file_path=file_path, change_type=self.change_type, diff=file_diff))
        return files

    def get_scannable_files(self, config: Config) -> Sequence[Union[DiffContentProvider, TextContentProvider]]:
        """Get files to scan. Output based on the `paths` field.

        Args:
            config: dict of credsweeper configuration

        Return:
            file objects for analysing

        """
        diff_data = self.load_patch_data(config)
        files = self.get_files_sequence(diff_data)
        return files



import re
from typing import Optional

from credsweeper.common import static_keyword_checklist
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueNotPartEncodedCheck(Filter):
    """Check that token is not a part of encoded data."""

    BASE64_ENCODED_DATA_PATTERN_BEFORE = re.compile(
        r"(^|[^A-Za-z0-9]+)(?P<val>(([A-Za-z0-9_-]{4}){16,64})|(([A-Za-z0-9+/]{4}){16,64}))([^=A-Za-z0-9]+|$)")
    BASE64_ENCODED_DATA_PATTERN_AFTER = re.compile(
        r"(^|[^A-Za-z0-9]+)(?P<val>(([A-Za-z0-9=_-]{4}){4,64})|(([A-Za-z0-9=+/]{4}){4,64}))([^=A-Za-z0-9]+|$)")

    def __init__(self, config: Config = None) -> None:
        pass

    @staticmethod
    def check_line_target_fit(line_data: LineData, target: AnalysisTarget) -> bool:
        """Verifies whether line data fit to be a part of many lines"""
        return line_data.line_num == target.line_num \
            and len(line_data.line) == target.line_len \
            and line_data.line == target.line \
            and 0 < target.line_num <= target.lines_len \
            and line_data.line == target.lines[target.line_num - 1]

    @staticmethod
    def check_val(line: str, pattern: re.Pattern) -> Optional[bool]:
        """Verifies whether the line looks like a base64 pattern"""
        if match_obj := pattern.match(line):
            val = match_obj.group("val")
            # not a path-like
            if not val.startswith('/') \
                    or not static_keyword_checklist.check_morphemes(val.lower(), 2) \
                    or '=' == val[-1]:
                # padding char is a marker too
                return True
        return None

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """

        if ValueNotPartEncodedCheck.check_line_target_fit(line_data, target):
            # suppose, there is plain lines order
            if 1 < target.line_num:
                result = ValueNotPartEncodedCheck.check_val(target.lines[line_data.line_num - 2],
                                                            ValueNotPartEncodedCheck.BASE64_ENCODED_DATA_PATTERN_BEFORE)
                if result is not None:
                    return result
            if target.lines_len > target.line_num:
                result = ValueNotPartEncodedCheck.check_val(target.lines[line_data.line_num],
                                                            ValueNotPartEncodedCheck.BASE64_ENCODED_DATA_PATTERN_AFTER)
                if result is not None:
                    return result
        else:
            # otherwise - need to iterate for all lines
            for i in range(target.lines_len):
                if line_data.line == target.lines[i]:
                    if 0 < i:
                        result = ValueNotPartEncodedCheck.check_val(
                            target.lines[i - 1], ValueNotPartEncodedCheck.BASE64_ENCODED_DATA_PATTERN_BEFORE)
                        if result is not None:
                            return result
                    i += 1
                    if target.lines_len > i:
                        result = ValueNotPartEncodedCheck.check_val(
                            target.lines[i], ValueNotPartEncodedCheck.BASE64_ENCODED_DATA_PATTERN_AFTER)
                        if result is not None:
                            return result
                    break
        return False



from credsweeper.filters.group.group import Group  # isort:skip

from credsweeper.filters.group.general_keyword import GeneralKeyword
from credsweeper.filters.group.general_pattern import GeneralPattern
from credsweeper.filters.group.password_keyword import PasswordKeyword
from credsweeper.filters.group.token_pattern import TokenPattern
from credsweeper.filters.group.url_credentials_group import UrlCredentialsGroup
from credsweeper.filters.group.weird_base36_token import WeirdBase36Token
from credsweeper.filters.group.weird_base64_token import WeirdBase64Token



import binascii
import contextlib

from credsweeper.common.constants import LATIN_1, ASCII
from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter
from credsweeper.utils import Util


class ValueAtlassianTokenCheck(Filter):
    """Check that candidate have a known structure"""

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received token which might be structured.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        value = line_data.value
        with contextlib.suppress(Exception):
            # atlassian integer:bytes from base64
            if value.startswith("BBDC-"):
                # Bitbucket HTTP Access Token
                return ValueAtlassianTokenCheck.check_atlassian_struct(value[5:])
            elif value.startswith("AT"):
                # Bitbucket App password
                while "\\=" in value or "%3d" in value or "%3D" in value:
                    # = sign may be escaped in URL https://www.rfc-editor.org/rfc/rfc3986
                    value = value.replace('\\', '')
                    value = value.replace('%3d', '=')
                    value = value.replace('%3D', '=')
                return ValueAtlassianTokenCheck.check_crc32_struct(value)
            else:
                # Jira / Confluence PAT token
                return ValueAtlassianTokenCheck.check_atlassian_struct(value)
        return True

    @staticmethod
    def check_crc32_struct(value: str) -> bool:
        """Returns False if value is valid for bitbucket app password structure 'payload:crc32'"""
        crc32 = int(value[-8:], 16)
        data = value[:-8].encode(ASCII)
        data_crc32 = binascii.crc32(data)
        if crc32 == data_crc32:
            return False
        return True

    @staticmethod
    def check_atlassian_struct(value: str) -> bool:
        """Returns False if value is valid for atlassian structure 'integer:bytes'"""
        decoded = Util.decode_base64(value, padding_safe=True, urlsafe_detect=True)
        delimiter_pos = decoded.find(b':')
        # there is limit for big integer value: math.log10(1<<64) = 19.265919722494797
        if 0 < delimiter_pos <= 20:
            val = decoded[:delimiter_pos].decode(LATIN_1)
            # at least 4 digits in the token
            if 1000 <= int(val):
                # test for ascii and Shannon entropy - there should be random data
                data = decoded[delimiter_pos + 1:]
                return Util.is_ascii_entropy_validate(data)
        return True



import re

from credsweeper.config import Config
from credsweeper.credentials import LineData
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.filters import Filter


class ValueTokenCheck(Filter):
    """Check if first substring of token is shorter than 5.

    Split candidate value into substrings using ` ;`{})(<>[]` separators. Check if first substring is shorter than 5

    Examples:
        "my password"
        "12);password"

    """

    SPLIT_PATTERN = r" |;|\)|\(|{|}|<|>|\[|\]|`"

    def __init__(self, config: Config = None) -> None:
        pass

    def run(self, line_data: LineData, target: AnalysisTarget) -> bool:
        """Run filter checks on received credential candidate data 'line_data'.

        Args:
            line_data: credential candidate data
            target: multiline target from which line data was obtained

        Return:
            True, if need to filter candidate and False if left

        """
        if line_data.is_well_quoted_value:
            return False
        tokens = re.split(self.SPLIT_PATTERN, line_data.value, maxsplit=1)
        # If tokens have length of 1 - pattern is not present in the value and original value returned from `.split(`
        if len(tokens) < 2:
            return False

        token = tokens[0]
        if len(token) < 5:
            return True

        return False



import logging
from abc import ABC, abstractmethod
from functools import cached_property
from typing import List, Optional, Generator

from credsweeper.common.constants import MAX_LINE_LENGTH
from credsweeper.file_handler.analysis_target import AnalysisTarget
from credsweeper.file_handler.descriptor import Descriptor
from credsweeper.utils import Util

logger = logging.getLogger(__name__)


class ContentProvider(ABC):
    """Base class to provide access to analysis targets for scanned object."""

    def __init__(
            self,  #
            file_path: Optional[str] = None,  #
            file_type: Optional[str] = None,  #
            info: Optional[str] = None) -> None:
        """
        Parameters:
            file_path: optional string. Might be specified if you know the file name where data were taken from.
            file_type: optional string. File extension e.g. ".java". It might be obtained from file_path if not given.
            info: optional string. Any information to help understand how a credential was found.

        """
        _file_path: str = file_path or ""
        _file_type: str = file_type if file_type is not None else Util.get_extension(file_path)
        _info: str = info or ""
        self.__descriptor = Descriptor(_file_path, _file_type, _info)

    @abstractmethod
    def yield_analysis_target(self, min_len: int) -> Generator[AnalysisTarget, None, None]:
        """Load and preprocess file diff data to scan.

        Args:
            min_len: minimal line length to scan

        Return:
            row objects to analysing

        """
        raise NotImplementedError()

    @cached_property
    def descriptor(self) -> Descriptor:
        """descriptor getter"""
        return self.__descriptor

    @cached_property
    def file_path(self) -> str:
        """file_path getter"""
        return self.__descriptor.path

    @cached_property
    def file_type(self) -> str:
        """file_type getter"""
        return self.__descriptor.extension

    @cached_property
    def info(self) -> str:
        """info getter"""
        return self.__descriptor.info

    @cached_property
    @abstractmethod
    def data(self) -> Optional[bytes]:
        """abstract data getter"""
        raise NotImplementedError(__name__)

    @abstractmethod
    def free(self) -> None:
        """free data after scan to reduce memory usage"""
        raise NotImplementedError(__name__)

    def lines_to_targets(
            self,  #
            min_len: int,
            lines: List[str],  #
            line_nums: Optional[List[int]] = None) -> Generator[AnalysisTarget, None, None]:
        """Creates list of targets with multiline concatenation"""
        lines_range = range(len(lines))
        if line_nums is None or len(line_nums) != len(lines):
            if line_nums is not None:
                logger.warning(
                    f"line numerations {len(line_nums)} does not match lines {len(lines)}. Plain numeration applied")
            line_nums = [1 + x for x in lines_range]

        for line_pos in lines_range:
            line = lines[line_pos]
            if min_len > len(line.strip()):
                # Ignore target if stripped part is too short for all types
                continue
            elif MAX_LINE_LENGTH < len(line):
                for chunk_start, chunk_end in Util.get_chunks(len(line)):
                    target = AnalysisTarget(
                        line_pos=line_pos,  #
                        lines=lines,  #
                        line_nums=line_nums,  #
                        descriptor=self.descriptor,  #
                        line=line[chunk_start:chunk_end],  #
                        offset=chunk_start)
                    yield target
            else:
                target = AnalysisTarget(line_pos, lines, line_nums, self.descriptor)
                yield target



import contextlib
import hashlib
import re
import string
from functools import cached_property
from typing import Any, Dict, Optional, Tuple

from colorama import Fore, Style

from credsweeper.common.constants import MAX_LINE_LENGTH, UTF_8, StartEnd, ML_HUNK
from credsweeper.config import Config
from credsweeper.utils import Util
from credsweeper.utils.entropy_validator import EntropyValidator


class LineData:
    """Object to treat and store scanned line related data.

    Parameters:
        key: Optional[str] = None
        line: string variable, line
        line_num: int variable, number of line in file
        path: string variable, path to file
        file_type: string variable, extension of file '.txt'
        info: additional info about how the data was detected
        pattern: regex pattern, detected pattern in line
        separator: optional string variable, separators between variable and value
        separator_start: optional variable, separator position start
        value: optional string variable, detected value in line
        variable: optional string variable, detected variable in line

    """

    quotation_marks = ('"', "'", '`')
    comment_starts = ("//", "* ", "#", "/*", "<!––", "%{", "%", "...", "(*", "--", "--[[", "#=")
    bash_param_split = re.compile("\\s+(\\-|\\||\\>|\\w+?\\>|\\&)")
    line_endings = re.compile(r"\\{1,8}[nr]")
    # https://en.wikipedia.org/wiki/Percent-encoding
    url_percent_split = re.compile(r"%(21|23|24|26|27|28|29|2a|2b|2c|2f|3a|3b|3d|3f|40|5b|5d)", flags=re.IGNORECASE)
    url_unicode_split = re.compile(r"\\u00(0000)?(21|23|24|26|27|28|29|2a|2b|2c|2f|3a|3b|3d|3f|40|5b|5d)",
                                   flags=re.IGNORECASE)
    # some symbols e.g. double quotes cannot be in URL string https://www.ietf.org/rfc/rfc1738.txt
    # \ - was added for case of url in escaped string \u0026amp; - means escaped & in HTML
    url_scheme_part_regex = re.compile(r"[0-9A-Za-z.-]{3}")
    url_chars_not_allowed_pattern = re.compile(r'[\s"<>\[\]^~`{|}]')
    url_value_pattern = re.compile(r'[^\s&;"<>\[\]^~`{|}]+[&;][^\s=;"<>\[\]^~`{|}]{3,80}=[^\s;&="<>\[\]^~`{|}]{1,80}')
    variable_strip_pattern = string.whitespace + """,'"-;"""

    INITIAL_WRONG_POSITION = -3
    EXCEPTION_POSITION = -2

    def __init__(
            self,  #
            config: Config,  #
            line: str,  #
            line_pos: int,  #
            line_num: int,  #
            path: str,  #
            file_type: str,  #
            info: str,  #
            pattern: re.Pattern,  #
            match_obj: Optional[re.Match] = None) -> None:
        self.config = config
        self.line: str = line
        self.line_pos: int = line_pos
        self.line_num: int = line_num
        self.path: str = path
        self.file_type: str = file_type
        self.info: str = info
        self.pattern: re.Pattern = pattern
        # do not store match object due it cannot be pickled with multiprocessing

        # start - end position of matched object
        self.value_start = LineData.INITIAL_WRONG_POSITION
        self.value_end = LineData.INITIAL_WRONG_POSITION
        self.key: Optional[str] = None
        self.separator: Optional[str] = None
        self.separator_start: int = LineData.INITIAL_WRONG_POSITION
        self.separator_end: int = LineData.INITIAL_WRONG_POSITION
        self.value: Optional[str] = None
        self.variable: Optional[str] = None
        self.variable_start = LineData.INITIAL_WRONG_POSITION
        self.variable_end = LineData.INITIAL_WRONG_POSITION
        self.value_leftquote: Optional[str] = None
        self.value_rightquote: Optional[str] = None
        # is set when variable & value are in URL for any source type
        self.url_part = False
        self.wrap = None
        self._3d_escaped_separator = False

        self.initialize(match_obj)

    def compare(self, other: 'LineData') -> bool:
        """Comparison method - skip whole line and checks only when variable and value are the same"""
        if self.path == other.path \
                and self.info == other.info \
                and self.line_num == other.line_num \
                and self.value_start == other.value_start \
                and self.variable == other.variable \
                and self.value == other.value:
            return True
        else:
            return False

    def initialize(self, match_obj: Optional[re.Match] = None) -> None:
        """Apply regex to the candidate line and set internal fields based on match."""
        if not isinstance(match_obj, re.Match) and isinstance(self.pattern, re.Pattern):
            match_obj = self.pattern.search(self.line, endpos=MAX_LINE_LENGTH)
        if match_obj is None:
            return

        def get_group_from_match_obj(_match_obj: re.Match, group: str) -> Any:
            with contextlib.suppress(Exception):
                return _match_obj.group(group)
            return None

        def get_span_from_match_obj(_match_obj: re.Match, group: str) -> Tuple[int, int]:
            with contextlib.suppress(Exception):
                span = _match_obj.span(group)
                return span[0], span[1]
            return LineData.EXCEPTION_POSITION, LineData.EXCEPTION_POSITION

        self.key = get_group_from_match_obj(match_obj, "keyword")
        self.separator = get_group_from_match_obj(match_obj, "separator")
        self.separator_start, self.separator_end = get_span_from_match_obj(match_obj, "separator")
        self.value = get_group_from_match_obj(match_obj, "value")
        self.value_start, self.value_end = get_span_from_match_obj(match_obj, "value")
        self.variable = get_group_from_match_obj(match_obj, "variable")
        self.variable_start, self.variable_end = get_span_from_match_obj(match_obj, "variable")
        self.value_leftquote = get_group_from_match_obj(match_obj, "value_leftquote")
        self.value_rightquote = get_group_from_match_obj(match_obj, "value_rightquote")
        self.wrap = get_group_from_match_obj(match_obj, "wrap")
        # percent encoded '=' in url
        self._3d_escaped_separator = bool(self.separator) and "%3D" == self.separator.upper()
        self.sanitize_value()
        self.sanitize_variable()

    def sanitize_value(self):
        """Clean found value from extra artifacts. Correct positions if changed."""
        if self.variable and self.value and not self.is_well_quoted_value:
            # sanitize is actual step for keyword pattern only
            _value = self.value
            self.clean_url_parameters()
            self.clean_bash_parameters()
            self.clean_toml_parameters()
            if 0 <= self.value_start and 0 <= self.value_end and len(self.value) < len(_value):
                start = _value.find(self.value)
                self.value_start += start
                self.value_end = self.value_start + len(self.value)

    def check_url_part(self) -> bool:
        """Determines whether value is part of url like line"""
        line_before_value = self.line[:self.value_start]
        url_pos = -1
        find_pos = 0
        while find_pos < self.value_start:
            # find rightmost pattern
            find_pos = line_before_value.find("://", find_pos)
            if -1 == find_pos:
                break
            else:
                url_pos = find_pos
                find_pos += 3
        # whether the line has url start pattern
        self.url_part = 3 <= url_pos
        self.url_part &= bool(self.url_scheme_part_regex.match(line_before_value, pos=url_pos - 3, endpos=url_pos))
        self.url_part &= not self.url_chars_not_allowed_pattern.search(line_before_value, pos=url_pos + 3)
        self.url_part |= self.line[self.variable_start - 1] in "?&" if 0 < self.variable_start else False
        self.url_part |= bool(self.url_value_pattern.match(self.value))
        self.url_part |= self._3d_escaped_separator
        return self.url_part

    def clean_url_parameters(self) -> None:
        """Clean url address from 'query parameters'.

        If line seem to be a URL - split by & character.
        Variable should be right most value after & or ? ([-1]). And value should be left most before & ([0])
        """
        if self.check_url_part():
            # all checks have passed - line before the value may be a URL
            self.variable = self.variable.rsplit('&')[-1].rsplit('?')[-1].rsplit(';')[-1]
            self.value = self.value.split('&', maxsplit=1)[0].split(';', maxsplit=1)[0].split('#', maxsplit=1)[0]
            if not self.variable.endswith("://"):
                # skip sanitize in case of URL credential rule
                self.value = self.url_unicode_split.split(self.value)[0]
                if self._3d_escaped_separator:
                    self.value = self.url_percent_split.split(self.value)[0]

    def clean_bash_parameters(self) -> None:
        """Split variable and value by bash special characters, if line assumed to be CLI command."""
        if self.variable.startswith("-"):
            value_spl = self.bash_param_split.split(self.value)
            # If variable name starts with `-` (usual case for args in CLI)
            #  and value can be split by bash special characters
            if len(value_spl) > 1:
                self.value = value_spl[0]
        if ' ' not in self.value and ("\\n" in self.value or "\\r" in self.value):
            value_whsp = self.line_endings.split(self.value)
            if len(value_whsp) > 1:
                self.value = value_whsp[0]

    def clean_toml_parameters(self) -> None:
        """Parenthesis, curly and squared brackets may be caught in TOML format and bash. Simple clearing"""
        cleaning_required = self.value and self.value[-1] in ['}', ']', ')']
        line_before_value = self.line[:self.value_start] if self.value_start and 0 <= self.value_start else ""
        while cleaning_required:
            cleaning_required = False
            for left, right in [('{', '}'), ('[', ']'), ('(', ')')]:
                if self.value.endswith(right) and left not in self.value \
                        and line_before_value.count(left) > line_before_value.count(right):
                    # full match does not reasonable to implement due open character may be in other line
                    self.value = self.value[:-1]
                    cleaning_required = True

    def sanitize_variable(self) -> None:
        """Remove trailing spaces, dashes and quotations around the variable. Correct position."""
        sanitized_var_len = 0
        variable = self.variable
        while self.variable and sanitized_var_len != len(self.variable):
            sanitized_var_len = len(self.variable)
            self.variable = self.variable.strip(self.variable_strip_pattern)
            if self.variable.endswith('\\'):
                self.variable = self.variable[:-1]
            if self.variable.startswith('{') and '}' in self.line[self.variable_end:]:
                # TOML case
                self.variable = self.variable[1:]
        if variable and len(self.variable) < len(variable) and 0 <= self.variable_start and 0 <= self.variable_end:
            start = variable.find(self.variable)
            self.variable_start += start
            self.variable_end = self.variable_start + len(self.variable)

    def is_comment(self) -> bool:
        """Check if line with credential is a comment.

        Return:
            True if line is a comment, False otherwise

        """
        cleaned_line = self.line.strip()
        for comment_start in self.comment_starts:
            if cleaned_line.startswith(comment_start):
                return True
        return False

    @cached_property
    def is_well_quoted_value(self) -> bool:
        """Well quoted value - means the value has been quoted or has line wrap"""
        result = False
        if self.value_leftquote and self.value_rightquote:
            if self.value_leftquote == self.value_rightquote:
                # regex caught well
                return True

            if 1 == len(self.value_leftquote):
                leftquote = self.value_leftquote
            else:
                # right side symbol should be a quote
                leftquote = self.value_leftquote[-1]
                if leftquote not in self.quotation_marks:
                    leftquote = ""

            if 1 == len(self.value_rightquote):
                rightquote = self.value_rightquote
            else:
                # clean \ sign in escaping text
                for q in self.value_rightquote:
                    if q in self.quotation_marks:
                        rightquote = q
                        break
                else:
                    rightquote = ""

            result = bool(leftquote) and (  #
                bool(rightquote) and (leftquote == rightquote)  # normal case
                or '\\' == self.value_rightquote and '\\' == self.line[-1]  # line wrap
            )

        elif self.value_leftquote:
            result = (  #
                ('\\' == self.value_rightquote or '\\' == self.value[-1]) and '\\' == self.line[-1]  # line wrap
                or '.php' == self.file_type  # php may use multiline string
                or 3 == self.value_leftquote.count('"') or 3 == self.value_leftquote.count("'")  # python multiline
            )

        return result

    @cached_property
    def is_quoted(self) -> bool:
        """Check if variable and value in a quoted string.

        Return:
            True if candidate in a quoted string, False otherwise

        """
        left_quote = None
        if 0 < self.variable_start:
            for i in self.line[:self.variable_start]:
                if i in ('"', "'", '`'):
                    left_quote = i
                    break
        right_quote = None
        if len(self.line) > self.value_end:
            for i in self.line[self.value_end:]:
                if i in ('"', "'", '`'):
                    right_quote = i
                    break
        result = bool(left_quote) and bool(right_quote) and left_quote == right_quote
        return result

    def is_source_file(self) -> bool:
        """Check if file with credential is a source code file or not (data, log, plain text).

        Return:
            True if file is source file, False otherwise

        """
        if not self.path:
            return False
        if Util.get_extension(self.path) in self.config.source_extensions:
            return True
        return False

    def is_source_file_with_quotes(self) -> bool:
        """Check if file with credential require quotation for string literals.

        Return:
            True if file require quotation, False otherwise

        """
        if not self.path:
            return False
        if Util.get_extension(self.path) in self.config.source_quote_ext:
            return True
        return False

    @staticmethod
    def get_hash_or_subtext(
            text: Optional[str],  #
            hashed: bool,  #
            cut_pos: Optional[StartEnd] = None,  #
    ) -> Optional[str]:
        """Represent not empty text with hash or a "beauty" subtext if required

        Args:
            text: str - input string
            hashed: bool - whether the text will be hashed and returned
            cut_pos: Optional[StartEnd] - start, end positions which text must be kept in output

        Return:
            sha256 hash in hex representation of input text with UTF-8 encodings
            or
            subtext from start to end, or original text as is

        """
        if text:
            if hashed:
                text = hashlib.sha256(text.encode(UTF_8, errors="strict")).hexdigest()
            elif cut_pos is not None:
                if 2 * ML_HUNK < cut_pos.end - cut_pos.start:
                    # subtext positions exceed the limit
                    text = text[cut_pos.start:cut_pos.end]
                else:
                    strip_text = text.strip()
                    if 2 * ML_HUNK >= len(strip_text):
                        # stripped text length meets the limit
                        text = strip_text
                    else:
                        offset = len(text) - len(text.lstrip())
                        center = (cut_pos.end + cut_pos.start - offset) >> 1
                        text = Util.subtext(strip_text, center, ML_HUNK)
        return text

    def to_str(self, subtext: bool = False, hashed: bool = False) -> str:
        """Represent line_data with subtext or|and hashed values"""
        cut_pos = StartEnd(self.variable_start, self.value_end) if subtext else None
        return f"line: '{self.get_hash_or_subtext(self.line, hashed, cut_pos)}'" \
               f" | line_num: {self.line_num} | path: {self.path}" \
               f" | value: '{self.get_hash_or_subtext(self.value, hashed)}'" \
               f" | entropy_validation: {EntropyValidator(self.value)}"

    def __str__(self):
        return self.to_str()

    def __repr__(self):
        return self.to_str(subtext=True)

    def to_json(self, hashed: bool, subtext: bool) -> Dict:
        """Convert line data object to dictionary.

        Return:
            Dictionary object generated from current line data

        """
        cut_pos = StartEnd(self.variable_start if 0 <= self.variable_start else self.value_start,
                           self.value_end) if subtext else None
        full_output = {
            "key": self.key,
            "line": self.get_hash_or_subtext(self.line, hashed, cut_pos),
            "line_num": self.line_num,
            "path": self.path,
            # info may contain variable name - so let it be hashed if requested
            "info": self.get_hash_or_subtext(self.info, hashed),
            "pattern": self.pattern.pattern,
            "separator": self.separator,
            "separator_start": self.separator_start,
            "separator_end": self.separator_end,
            "value": self.get_hash_or_subtext(self.value, hashed),
            "value_start": self.value_start,
            "value_end": self.value_end,
            "variable": self.get_hash_or_subtext(self.variable, hashed),
            "variable_start": self.variable_start,
            "variable_end": self.variable_end,
            "value_leftquote": self.value_leftquote,
            "value_rightquote": self.value_rightquote,
            "entropy_validation": EntropyValidator(self.value).to_dict()
        }
        reported_output = {k: v for k, v in full_output.items() if k in self.config.line_data_output}
        return reported_output

    def get_colored_line(self, hashed: bool, subtext: bool = False) -> str:
        """Represents the LineData with a value, separator, and variable color formatting"""
        if hashed:
            # return colored hash
            return Fore.LIGHTGREEN_EX \
                + self.get_hash_or_subtext(self.line, hashed,
                                           StartEnd(self.value_start, self.value_end) if subtext else None) \
                + Style.RESET_ALL
        # at least, value must present
        line = self.line[:self.value_start] \
               + Fore.LIGHTYELLOW_EX \
               + self.line[self.value_start:self.value_end] \
               + Style.RESET_ALL \
               + self.line[self.value_end:]  # noqa: E127
        # separator may be missing
        if 0 <= self.separator_start < self.separator_end <= self.value_start:
            line = line[:self.separator_start] \
                   + Fore.LIGHTGREEN_EX \
                   + line[self.separator_start:self.separator_end] \
                   + Style.RESET_ALL \
                   + line[self.separator_end:]
        # variable may be missing
        if 0 <= self.separator_start \
                and 0 <= self.variable_start < self.variable_end <= self.separator_end <= self.value_start \
                or 0 <= self.variable_start < self.variable_end <= self.value_start:
            line = line[:self.variable_start] \
                   + Fore.LIGHTBLUE_EX \
                   + line[self.variable_start:self.variable_end] \
                   + Style.RESET_ALL \
                   + line[self.variable_end:]
        if subtext:
            # display part of the text, centered around the start of the value, style reset at the end as a fallback
            line = f"{Util.subtext(line, self.value_start + len(line) - len(self.line), ML_HUNK)}{Style.RESET_ALL}"
        return line



from credsweeper.credentials.candidate import Candidate
from credsweeper.credentials.candidate_group_generator import CandidateGroupGenerator
from credsweeper.credentials.candidate_key import CandidateKey
from credsweeper.credentials.credential_manager import CredentialManager
from credsweeper.credentials.line_data import LineData



from credsweeper.app import CredSweeper
from credsweeper.common.constants import ThresholdPreset
from credsweeper.file_handler import ContentProvider, ByteContentProvider, DiffContentProvider, StringContentProvider, \
    DataContentProvider, \
    TextContentProvider
from credsweeper.ml_model.ml_validator import MlValidator

__all__ = [
    'ByteContentProvider',  #
    'ContentProvider',  #
    'CredSweeper',  #
    'DataContentProvider',  #
    'DiffContentProvider',  #
    'MlValidator',  #
    'StringContentProvider',  #
    'TextContentProvider',  #
    'ThresholdPreset',  #
    '__version__'
]

__version__ = "1.10.5"
